#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["compute_conv_output", "compute_onnx_conv_output"]

from .base_layer import OnnxLayer
from ..graph_tools import get_field


def compute_conv_output(in_shape, kernel_shape, strides, pads=None):
    """Compute the output shape after applying a convolution for each spatial dimension

    Args:
        in_shape (tuple of ints): the input shape.
        kernel_shape (tuple of ints): the convolutional kernel shape.
        strides (tuple of ints): the convolutional strides.
        pads (tuple of ints): the pads to apply. Defaults to None.

    Returns:
        tuple of ints: the output shape after applying the convolution.
    """
    N = len(in_shape)
    assert N == len(strides) == len(kernel_shape)
    if pads is None:
        pads = [0] * N * 2
    assert len(in_shape) == len(pads) / 2

    out_shape = []
    for x, kx, sx, pl, pr in zip(in_shape, kernel_shape, strides, pads[:N], pads[N:]):
        y = (x + pl + pr - kx) // sx + 1
        out_shape.append(y)
    return tuple(out_shape)


def compute_onnx_conv_output(qlayer, input_shape, apply_pool=True, transpose=False):
    """Compute the output shape produced by qlayer for a given input shape.

    Args:
        qlayer (OnnxLayer): the layer with the parameters.
        input_shape (tuple of ints): the input shape.
        apply_pool (bool, optional): whether output shape has to include pool operation.
            Defaults to True.
        transpose (bool, optional): whether the input performs the transpose of the
            convolution. Defaults to False.

    Returns:
        tuple of ints: the output shape after applying the convolution.
    """
    isinstance(qlayer, OnnxLayer)
    kernel = qlayer.weights["kernel"]

    # Compute conv output shape
    spatial_dims = kernel.ndim - 2
    conv_strides = get_field(qlayer, "strides", (1,) * spatial_dims)

    if transpose:
        # Because of weight ordering (CFHW), the operation is different for transpose convolutions.
        output_shape = compute_transpose_conv_output(
            input_shape[2:], kernel.shape[2:], conv_strides)
        groups = get_field(qlayer, "groups", 1)
        n_filters = (kernel.shape[1] * groups,)
        return input_shape[:1] + n_filters + output_shape
    else:
        conv_pads = qlayer.weights["pads"].tolist()
        if any(conv_pads[:2]) or any(conv_pads[kernel.ndim:kernel.ndim + 2]):
            raise ValueError(f"Unsupported pads format in {qlayer.name}: no pad can be applied "
                             f"in batch or spatial dimension. Receives: {conv_pads}")
        conv_pads = conv_pads[2:kernel.ndim] + conv_pads[kernel.ndim + 2:]
        output_shape = compute_conv_output(
            input_shape[2:], kernel.shape[2:], conv_strides, conv_pads)

    # Compute max pool output shape
    if apply_pool:
        if "MaxPool" in qlayer.op_type:
            pool_size = get_field(qlayer, "pool_size", (2,) * spatial_dims)
            pool_strides = get_field(qlayer, "pool_strides", (1,) * spatial_dims)
            pool_pads = get_field(qlayer, "pool_pads", (0, 0) * spatial_dims)
            output_shape = compute_conv_output(output_shape, pool_size, pool_strides, pool_pads)
        elif "GlobalAvgPool" in qlayer.op_type:
            output_shape = (1,) * spatial_dims
    return input_shape[:1] + kernel.shape[:1] + output_shape


def compute_transpose_conv_output(in_shape, kernel_shape, strides):
    """Compute the output shape after applying a transpose convolution
    on each spatial dimension

    Args:
        in_shape (tuple of ints): the input shape.
        kernel_shape (tuple of ints): the convolutional kernel shape.
        strides (tuple of ints): the convolutional strides.

    Returns:
        tuple of ints: the output shape after applying the transpose convolution.
    """
    N = len(in_shape)
    assert N == len(strides) == len(kernel_shape)

    out_shape = []
    for x, kx, sx in zip(in_shape, kernel_shape, strides):
        # simplified from torch computation (removing dilation and padding)
        # dim_out = stride[i] * (dim_in - 1) + kernel_shape[i]
        y = sx * (x - 1) + kx
        out_shape.append(y)
    return tuple(out_shape)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["convert_min_max_to_clip"]

import numpy as np
import onnx
import onnx.numpy_helper

from ..model import ONNXModel


def convert_min_max_to_clip(model):
    """Transforms Min or Max nodes into a Clip node.

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    assert isinstance(model, ONNXModel)

    def _swap_inputs(model, node):
        # for Min and Max node, we need to put the non initializer input on the first
        # position if it is not.
        initializer = model.get_initializer(node.input[0])
        if initializer is not None:
            node.input[0], node.input[1] = node.input[1], node.input[0]

    for node in model.nodes():
        # Min or Max nodes should have exactly two inputs
        if node.op_type in ("Max", "Min") and len(node.input) == 2:
            # if both inputs are non initializers, we do nothing
            if len(model.get_node_inputs(node)) == 2:
                return

            _swap_inputs(model, node)

            node_value = model.get_variable(node.input[1])
            # Conversion from Min/Max to Clip is not possible when the node's value
            # size is greater than 1, as Clip only supports a scalar value.
            if node_value.size != 1:
                continue

            node_value_tp = model.get_initializer(node.input[1])
            node_value_tp.CopyFrom(onnx.numpy_helper.from_array(
                np.array(node_value.item(), np.float32), node.input[1]))

            if node.op_type == "Min":
                # Add an empty min. Min should be before Max that's why we swap min and max
                node.input.append("")
                node.input[1], node.input[2] = node.input[2], node.input[1]
            # convert to Clip
            node.op_type = "Clip"
    model.check_model()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["StatefulRecurrent", "QuantizedStatefulRecurrent", "reset_states",
           "StatefulProjection", "QuantizedStatefulProjection"]

import keras
import tensorflow as tf

from .recorders import NonTrackVariable, TensorRecorder, NonTrackFixedPointVariable
from .layers_base import (register_quantize_target, tensor_inputs, apply_buffer_bitwidth,
                          register_aligned_inputs, QuantizedLayer, neural_layer_init,
                          rescale_outputs)
from .quantizers import WeightQuantizer, OutputQuantizer
from ..tensors import FixedPoint, QTensor
from ..debugging import assert_equal, assert_less


@keras.saving.register_keras_serializable()
class StatefulRecurrent(keras.layers.Layer):
    """ A recurrent layer with an internal state.

    Args:
        subsample_ratio (float, optional): subsampling ratio that defines rate at which outputs are
            produced (zero otherwise). Defaults to 1.
    """

    def __init__(self, *args, subsample_ratio=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.subsample_ratio = subsample_ratio
        self._counter = NonTrackVariable("counter")
        self._internal_state_real = NonTrackVariable("internal_state_real")
        self._internal_state_imag = NonTrackVariable("internal_state_imag")

    @property
    def counter(self):
        return self._counter.var

    @property
    def internal_state_real(self):
        return self._internal_state_real.var

    @property
    def internal_state_imag(self):
        return self._internal_state_imag.var

    def build(self, input_shape):
        with tf.name_scope(self.name + '/'):
            super().build(input_shape)
            # 'A' weight is a complex64 tensor stored as two float32 tensor to ease quantization
            self.A_real = self.add_weight(name='A_real', shape=(input_shape[-1],))
            self.A_imag = self.add_weight(name='A_imag', shape=(input_shape[-1],))

    def update_counter(self):
        """Function that increments a 'counter' variable. If the variable doesn't exist, it
            will be created in the layer graph and tracked.
        """
        self._counter.init_var(0, True)
        self._counter.var.assign_add(1)

    def call(self, inputs):
        """ This call method only takes in a single input step.

        For every input step, the internal state is updated using the inputs which should be the
        updated state from the previous layer.
        """

        # Update or initialize (if it's the first call) the counter variable
        self.update_counter()

        # Initialize (only during the first call) the internal state variables
        self._internal_state_real.init_var(inputs)
        self._internal_state_imag.init_var(inputs)

        # Since the above variables will be added to the graph during the layer graph initialization
        # the batch size is unknown and set to None. Such behavior won't allow the next
        # operations due to a conflict in the brodcastable shape. As a workaround, we set again in
        # the variables a tf.zeros tensor but this time with the batch_size known since it's in the
        # call, when counter == 1 (first passage or after a reset, when the batch_size changes for
        # example).
        if self.counter == 1:
            self._internal_state_real.set_var(tf.zeros_like(inputs))
            self._internal_state_imag.set_var(tf.zeros_like(inputs))

        # Path for skipping computations
        if self.subsample_ratio and tf.math.floormod(self.counter, self.subsample_ratio) != 0:
            return tf.zeros_like(tf.stack([self.internal_state_real, self.internal_state_imag],
                                          -1))

        # Update internal state: compute real and imaginary part separately
        updated_real = self.internal_state_real * self.A_real - \
            self.internal_state_imag * self.A_imag + inputs
        internal_state_imag = self.A_imag * self.internal_state_real + \
            self.A_real * self.internal_state_imag
        self._internal_state_imag.set_var(internal_state_imag)
        # Update real part in a second time so that it does not impact imaginary part computation
        self._internal_state_real.set_var(updated_real)

        return tf.stack([self.internal_state_real, self.internal_state_imag], -1)

    def get_config(self):
        config = super().get_config()
        config.update({
            'subsample_ratio': self.subsample_ratio
        })
        return config

    def reset_layer_states(self):
        for var in [self._counter, self._internal_state_real, self._internal_state_imag]:
            var.reset_var()


@register_quantize_target([StatefulRecurrent])
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedStatefulRecurrent(QuantizedLayer, StatefulRecurrent):
    """ A quantized version of the StatefulRecurrent layer that operates on quantized inputs,
    weights and internal state.

    Note that internal state is quantized to 16-bits for accuracy reasons, inputs and outputs of
    this layer are then also 16-bits.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """

    def __init__(self, *args, quant_config=None, **kwargs):
        super().__init__(*args, quant_config=quant_config, **kwargs)

        self._internal_state_real = NonTrackFixedPointVariable("internal_state_real")
        self._internal_state_imag = NonTrackFixedPointVariable("internal_state_imag")

        # Build weight quantizer for A_real and A_imag (sharing the same quantizer)
        if "a_quantizer" not in self.quant_config:
            # Forcing to:
            #   - per-tensor to ensure alignement in the call operations
            #   - 16-bits for accuracy reasons
            #   - FixedPoint quantization to prevent scale_out operations on internal_state
            self.quant_config["a_quantizer"] = {"bitwidth": 16, "axis": None, "fp_quantizer": True}
        a_quantizer_cfg = self.quant_config["a_quantizer"]
        self.a_quantizer = WeightQuantizer(name="a_quantizer", **a_quantizer_cfg)

        # Finalize output quantizer, add one with default configuration if there is None in the
        # config as state must be quantized
        if "output_quantizer" not in self.quant_config:
            self.quant_config["output_quantizer"] = {"bitwidth": 16, "axis": "per-tensor"}
        out_quant_cfg = self.quant_config["output_quantizer"]
        self.out_quantizer = OutputQuantizer(name="output_quantizer", **out_quant_cfg)
        self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=True)

        # Prepare the variable that should be recorded
        self.input_add_shift = TensorRecorder(name=self.name + "/input_add_shift")

    @tensor_inputs([FixedPoint])
    def call(self, inputs):
        # Update or initialize (if it's the first call) the counter variable
        self.update_counter()

        # Initialize (only during the first call) the internal state variables
        self._internal_state_real.init_var(inputs)
        self._internal_state_imag.init_var(inputs)

        # Set again the variables with a fully defined shape (batch_size is known)
        if self.counter == 1:
            self._internal_state_real.set_var(tf.zeros_like(inputs))
            self._internal_state_imag.set_var(tf.zeros_like(inputs))

        # Promote internal_state
        internal_state_real = self._internal_state_real.promote(self.buffer_bitwidth)
        internal_state_imag = self._internal_state_imag.promote(self.buffer_bitwidth)

        if self.subsample_ratio and tf.math.floormod(self.counter, self.subsample_ratio) != 0:
            # Path for skipping computations
            return self.out_quantizer(tf.zeros_like(tf.stack([internal_state_real,
                                                              internal_state_imag], -1)))

        # Quantize A matrices
        A_real = self.a_quantizer(self.A_real)
        A_imag = self.a_quantizer(self.A_imag)

        # The op below is possible because internal_state is quantized per-tensor on the real/imag
        # concatenated tensor, thus internal_state_real and internal_state_imag are aligned.
        assert_equal(internal_state_real.frac_bits, internal_state_imag.frac_bits)

        # Update internal state: compute real and imaginary part separately
        updated_real = tf.multiply(internal_state_real, A_real) - \
            tf.multiply(internal_state_imag, A_imag)

        # Align inputs with intermediate result before addition
        # By construction, A_real² + A_imag² < 1 will ensure that A matrices are between 0 and 1
        # and thus their FixedPoint representation is all fractional part and this will also be True
        # for updated_state. So the higher frac_bits are on the state (input_proj being 16-bits,
        # frac_bits it at most 15).
        assert_less(inputs.frac_bits, updated_real.frac_bits), f"Input resolution is to high \
            {inputs.frac_bits} must be below {updated_real.frac_bits}."
        inputs, input_add_shift = inputs.align(updated_real, self.buffer_bitwidth)
        self.input_add_shift(input_add_shift)

        # At this point addition is possible
        updated_real = updated_real + inputs

        # Same for imaginary part
        updated_imag = tf.multiply(internal_state_real, A_imag) + \
            tf.multiply(internal_state_imag, A_real)

        # Concatenate updated_real and updated_imag
        concatenated_internal_state = tf.stack([updated_real, updated_imag], -1)

        # Quantize down the internal state
        output = self.out_quantizer(concatenated_internal_state)

        # Update internal state members, storying float state
        self._internal_state_real.set_var(tf.gather(output, 0, axis=-1))
        self._internal_state_imag.set_var(tf.gather(output, 1, axis=-1))
        return output


def reset_states(model):
    """ Resets all StatefulRecurrent layers internal states in the model.
    Args:
        model (keras.Model): the model to reset
    """
    for layer in model.layers:
        if isinstance(layer, StatefulRecurrent):
            layer.reset_layer_states()


@keras.saving.register_keras_serializable()
class StatefulProjection(keras.layers.Dense):
    """ Same as a Dense layer but with optional reshaping operation.

    Reshaping can happen both on the inputs and the outputs.

    Args:
        downshape (tuple, optional): target shape for downshape operation that happens before the
            dense. Defaults to None.
        upshape (tuple, optional): target shape for upshape operation that happens after the dense.
            Defaults to None.
    """

    def __init__(self, *args, downshape=None, upshape=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.downshape = tuple(downshape) if downshape else None
        self.upshape = tuple(upshape) if upshape else None

    def build(self, input_shape):
        with tf.name_scope(self.name + '/'):
            if self.downshape is None:
                super().build(input_shape)
            else:
                # When downshape is enabled, build the layer with the target shape so that variables
                # are build properly
                super().build(self.downshape)
                # Edit the input_spec so that from a graph point of view, this layer sees the
                # original input shape
                last_dim = tf.TensorShape(input_shape)[-1]
                self.input_spec = tf.keras.layers.InputSpec(min_ndim=2, axes={-1: last_dim})

    def call(self, inputs):
        # Apply the optional input downshape
        if self.downshape is not None:
            inputs = tf.reshape(inputs, (tf.shape(inputs)[0],) + self.downshape)

        # Standard Dense operation
        outputs = super().call(inputs)

        # Apply the optional output upshape
        if self.upshape is not None:
            outputs = tf.reshape(outputs, (tf.shape(outputs)[0],) + self.upshape)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["downshape"] = self.downshape
        config["upshape"] = self.upshape
        return config


@register_quantize_target([StatefulProjection])
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedStatefulProjection(QuantizedLayer, StatefulProjection):
    """ A quantized version of the StatefulProjection layer that operates on quantized inputs.
    """
    @neural_layer_init(False)
    def __init__(self, *args, **kwargs):
        # Limit buffer bitwidth to 27 for HW constraint
        self.quant_config['buffer_bitwidth'] = min(28, self.quant_config['buffer_bitwidth'])
        self.buffer_bitwidth = self.quant_config['buffer_bitwidth'] - 1

        # Weight quantizer must be per-tensor to allow upshaping, override it when necessary
        if self.upshape is not None:
            self.quant_config["weight_quantizer"]["axis"] = None
            weight_quantizer_cfg = self.quant_config["weight_quantizer"]
            self.weight_quantizer = WeightQuantizer(name="weight_quantizer",
                                                    **weight_quantizer_cfg)

    @tensor_inputs([QTensor, tf.Tensor])
    @rescale_outputs
    def call(self, inputs):
        if self.downshape is not None:
            inputs = tf.reshape(inputs, (tf.shape(inputs)[0],) + self.downshape)

        # Quantize the weights
        kernel = self.weight_quantizer(self.kernel)

        outputs = tf.matmul(inputs, kernel)

        if self.use_bias:
            # Quantize and align biases
            bias = self.bias_quantizer(self.bias, outputs)
            outputs = tf.add(outputs, bias)

        if self.upshape is not None:
            outputs = tf.reshape(outputs, (tf.shape(outputs)[0],) + self.upshape)
        return outputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["PaddedConv2D", "QuantizedConv2D", "QuantizedConv2DTranspose"]

import tensorflow as tf
import numpy as np
import keras

from keras import layers, backend

from .layers_base import (register_quantize_target, rescale_outputs, tensor_inputs,
                          neural_layer_init, register_aligned_inputs, QuantizedLayer)
from .quantizers import AlignedWeightQuantizer
from ..tensors import QTensor, QFloat, FixedPoint


def apply_padding(inputs, strides, kernel_size, padding_values):
    """Apply "SAME" padding to the inputs

    Args:
        inputs (:obj:`QFloat`, tf.Tensor): the inputs tensor.
        strides (tuple): the strides tuple.
        kernel_size (int): the kernel size.
        padding_values (:obj:`QFloat`, tf.Tensor): the padding values to apply.

    Returns:
        tf.Tensor or :obj:`QFloat`: padded inputs.
    """
    _, h, w, _ = inputs.shape
    filter_width = kernel_size[0]
    filter_height = kernel_size[1]
    if h % strides[0] == 0:
        pad_along_height = max(filter_height - strides[0], 0)
    else:
        pad_along_height = max(filter_height - (h % strides[0]), 0)
    if w % strides[1] == 0:
        pad_along_width = max(filter_width - strides[1], 0)
    else:
        pad_along_width = max(filter_width - (w % strides[1]), 0)
    pad_top = pad_along_height // 2
    pad_bottom = pad_along_height - pad_top
    pad_left = pad_along_width // 2
    pad_right = pad_along_width - pad_left
    padding = [[0, 0], [pad_top, pad_bottom], [pad_left, pad_right], [0, 0]]

    if padding_values.shape.ndims == 0:
        # per-tensor algorithm is faster than per-axis one
        output = tf.pad(inputs, padding, "CONSTANT", padding_values)
    else:
        # The per-channel algorithm consists of:
        # - subtract the padding value to inputs
        # - add a zero padding
        # - add the padding value to revert the subtraction from inputs and add the desired value
        #   around
        inputs = inputs - padding_values
        zero_pad = 0.
        if isinstance(inputs, QFloat):
            zero_pad = QFloat(FixedPoint(0, inputs.value_bits, inputs.frac_bits), inputs.scales)
        output = tf.pad(inputs, padding, "CONSTANT", zero_pad)
        output = output + padding_values

    return output


def deconv_output_length(input_length, filter_size, padding,
                         output_padding=None, stride=0, dilation=1):
    """Determines output length of a transposed convolution given input length.

    Args:
        input_length (int): length of input.
        filter_size (int): size of the kernel.
        padding (str): one of `"same"`, `"valid"`, `"full"`.
        output_padding (int, optional): amount of padding along the output dimension. Can be set to
            `None` in which case the output length is inferred. Defaults to None.
        stride (int, optional): stride value. Defaults to 0.
        dilation (int, optional): dilation value. Defaults to 1.

    Returns:
        int: the output length
    """
    assert padding in {"same", "valid", "full"}
    if input_length is None:
        return None

    # Get the dilated kernel size
    filter_size = filter_size + (filter_size - 1) * (dilation - 1)

    # Infer length if output padding is None, else compute the exact length
    if output_padding is None:
        if padding == "valid":
            length = input_length * stride + max(filter_size - stride, 0)
        elif padding == "full":
            length = input_length * stride - (stride + filter_size - 2)
        elif padding == "same":
            length = input_length * stride

    else:
        if padding == "same":
            pad = filter_size // 2
        elif padding == "valid":
            pad = 0
        elif padding == "full":
            pad = filter_size - 1

        length = ((input_length - 1) * stride + filter_size - 2 * pad + output_padding)
    return length


@keras.saving.register_keras_serializable()
class PaddedConv2D(layers.Conv2D):
    """A convolutional layer that can use custom padding values.

    Note that when padding values are provided, padding 'SAME' will be applied with the provided
    value (overriding 'padding' parameter).

    Args:
        padding_value (float, list, tensor, optional): the value or the list of values used when
            padding for the 'same' convolution type. Padding is per-tensor if one value is provided
            or per-channel otherwise. If None, zero-padding is used. Defaults to None.
    """

    def __init__(self, *args, padding_value=None, **kwargs):
        super().__init__(*args, **kwargs)
        if padding_value is not None:
            try:
                self._padding_value = list(padding_value)
            except TypeError:
                self._padding_value = [padding_value]
            # When a custom padding_value is given, self.padding is overwritten and 'SAME' padding
            # will be done explicitly in the call.
            self.padding = 'valid'
        else:
            self._padding_value = None

    def build(self, input_shape):
        # Ensure variables are build with the appropriate name, this is required for usage with
        # Model.from_config that happens in sanitize
        with tf.name_scope(self.name + '/'):
            super().build(input_shape)

    def call(self, inputs):
        # We need a custom padding for specifics padding values
        if self._padding_value is not None:
            # Raise an error if the padding values have not the expected number of values
            pad_size = np.size(self._padding_value)
            if pad_size not in (1, inputs.shape[-1]):
                raise ValueError("The padding value must be a scalar or a list of values, for "
                                 f"a value per input channel. Receives {self._padding_value}.")
            t_padding = tf.cast(self._padding_value, dtype=tf.float32)
            # Note that we use the custom padding even when padding value provided is 0.
            inputs = apply_padding(inputs,
                                   list(self.strides),
                                   (self.kernel.shape[1], self.kernel.shape[0]),
                                   t_padding)

        outputs = self.convolution_op(inputs, self.kernel)

        if self.use_bias:
            outputs = tf.add(outputs, self.bias)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["padding_value"] = self._padding_value
        return config


@register_quantize_target([layers.Conv2D, PaddedConv2D])
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedConv2D(QuantizedLayer, layers.Conv2D):
    """A convolutional layer that operates on quantized inputs and weights.

    Note that when padding values are provided, padding 'SAME' will be applied with the provided
    value (overriding 'padding' parameter).

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
        padding_value (float, list, tensor, optional): the value or the list of values used when
            padding for the 'same' convolution type. Padding is per-tensor if one value is provided
            or per-channel otherwise. If None, zero-padding is used. Defaults to None.
    """
    @neural_layer_init(False)
    def __init__(self, *args, padding_value=None, quant_config=None, **kwargs):
        # Limit buffer bitwidth to 27
        self.quant_config['buffer_bitwidth'] = min(28, self.quant_config['buffer_bitwidth'])
        self.buffer_bitwidth = self.quant_config['buffer_bitwidth'] - 1

        if padding_value is not None:
            try:
                self.padding_value = list(padding_value)
            except TypeError:
                self.padding_value = [padding_value]
            # We quantize the padding_value with the same bitwidth as the weights
            self.padding_value_quantizer = AlignedWeightQuantizer(self.weight_quantizer.bitwidth,
                                                                  signed=True)
            # When a custom padding_value is given, self.padding is overwritten and 'SAME' padding
            # will be done explicitely in the call.
            self.padding = 'valid'
        else:
            self.padding_value = None

    @tensor_inputs([QTensor, tf.Tensor])
    @rescale_outputs
    def call(self, inputs):
        # Quantize the weights
        kernel = self.weight_quantizer(self.kernel)

        if self.padding_value is not None:
            # We need to align the padding value on the inputs, which is only supported for QFloat
            # Note that this restriction is purely to reduce the cognitive and test cost: should
            # we need at some point to support it, we could extend the AlignedWeightQuantizer or
            # wrap the inputs in a QFloat of scale 1.
            if not isinstance(inputs, QFloat):
                raise ValueError("When using a non-zero padding value, the inputs must be QFloat.")

            pad_size = np.size(self.padding_value)
            if pad_size not in (1, inputs.shape[-1]):
                raise ValueError("The padding value must be a scalar or a list of values, for "
                                 f"a value per input channel. Receives {self.padding_value}.")
            t_padding_value = tf.cast(self.padding_value, dtype=tf.float32)
            # Quantize and align padding_value on the inputs
            q_padding_value = self.padding_value_quantizer(t_padding_value, inputs)

            # Note that we use the custom padding even when padding value provided is 0.
            inputs = apply_padding(inputs,
                                   list(self.strides),
                                   (kernel.shape[1], kernel.shape[0]),
                                   q_padding_value)

        if isinstance(self.padding, str):
            tf_padding = self.padding.upper()
        else:
            tf_padding = self.padding

        outputs = tf.nn.convolution(inputs,
                                    kernel,
                                    strides=list(self.strides),
                                    padding=tf_padding,
                                    dilations=list(self.dilation_rate),
                                    data_format=self._tf_data_format,
                                    name=self.__class__.__name__)

        if self.use_bias:
            # Quantize bias and align it on the outputs
            bias = self.bias_quantizer(self.bias, outputs)
            outputs = tf.add(outputs, bias)

        return outputs

    def get_config(self):
        config = super().get_config()
        if self.padding_value is not None:
            config["padding_value"] = self.padding_value
        return config


@register_quantize_target(layers.Conv2DTranspose)
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedConv2DTranspose(QuantizedLayer, layers.Conv2DTranspose):
    """A transposed convolutional layer that operates on quantized inputs and weights.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """
    @neural_layer_init(False)
    def __init__(self, *args, quant_config=None, **kwargs):
        # Limit buffer bitwidth to 27
        self.quant_config['buffer_bitwidth'] = min(28, self.quant_config['buffer_bitwidth'])
        self.buffer_bitwidth = self.quant_config['buffer_bitwidth'] - 1

    @tensor_inputs([FixedPoint])
    @rescale_outputs
    def call(self, inputs):
        # Quantize the weights
        # Conv2DTranspose weight shape is (h, w, filters, channels) which cannot be quantized
        # 'per-axis' as it is: the last dimension is not the output last dimension (i.e filters)
        # so it must be transposed to (h, w, channels, filters) before quantization to have the
        # appropriate frac_bits shape. Kernel values will then be transposed back to apply the
        # Tensorflow operation.
        kernel = self.weight_quantizer(tf.transpose(self.kernel, (0, 1, 3, 2)))

        # Prepare deconvolution output shape
        inputs_shape = tf.shape(inputs)
        batch_size, height, width = inputs_shape[0], inputs_shape[1], inputs_shape[2]
        kernel_h, kernel_w = self.kernel_size
        stride_h, stride_w = self.strides

        if self.output_padding is None:
            out_pad_h = out_pad_w = None
        else:
            out_pad_h, out_pad_w = self.output_padding

        # Infer the dynamic output shape
        out_height = deconv_output_length(height,
                                          kernel_h,
                                          padding=self.padding,
                                          output_padding=out_pad_h,
                                          stride=stride_h,
                                          dilation=self.dilation_rate[0])
        out_width = deconv_output_length(width,
                                         kernel_w,
                                         padding=self.padding,
                                         output_padding=out_pad_w,
                                         stride=stride_w,
                                         dilation=self.dilation_rate[1])

        output_shape = (batch_size, out_height, out_width, self.filters)
        output_shape_tensor = tf.stack(output_shape)

        # Do transposed convolution
        outputs = backend.conv2d_transpose(inputs, kernel, output_shape_tensor,
                                           self.strides, self.padding, self.data_format,
                                           self.dilation_rate)

        if self.use_bias:
            # Quantize biases and align them on the inputs
            bias = self.bias_quantizer(self.bias, outputs)
            # Add biases
            outputs = tf.add(outputs, bias)

        return outputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tempfile
from pathlib import Path

import onnx
from onnxruntime.quantization import create_calibrator, CalibrationMethod

from .register_patterns import PATTERNS_MAP, CUSTOM_PATTERNS_MAP
from .data_reader import CalibrationDataReader


def _get_op_types_to_calibrate():
    # This function computes the set of operation types whose outputs need to be calibrated.
    # These operation types are the last operations in each pattern from
    # PATTERNS_MAP and CUSTOM_PATTERNS_MAP.
    return {pattern.pattern[-1] for pattern in PATTERNS_MAP + CUSTOM_PATTERNS_MAP}


def calibrate(model,
              samples=None,
              num_samples=None,
              batch_size=None,
              symmetric=False,
              average=False,
              per_tensor_activations=True):
    """Calibrates the ONNX model using the provided samples.

    When no samples are provided, random samples are generated.

    Args:
        model (ModelProto): onnx model to calibrate
        samples (np.array, optional): calibration samples. When no samples are provided,
            random samples are generated. Defaults to None.
        num_samples (int, optional): number of samples to generate. Defaults to None.
        batch_size (int, optional): the batch size. Defaults to None.
        symmetric (bool, optional): whether the final range of tensor during calibration
            will be explicitly set to symmetric to central point "0". Defaults to False.
        average (bool, optional): whether average of the minimum and maximum values
            will be computed. Defaults to False.
        per_tensor_activations (bool, optional): wheter to compute activation ranges per tensor.
            Defaults to True.

    Returns:
        dict: tensor names with calibration ranges.
    """
    # Create a calibration data reader from given samples.
    calibration_data_reader = CalibrationDataReader(model, samples, num_samples, batch_size)

    with tempfile.TemporaryDirectory(prefix="ort.quant.") as quant_tmp_dir:
        tmp_dir = Path(quant_tmp_dir)

        # Temporary save the model to create the calibrator
        model_path = tmp_dir.joinpath("model.onnx")
        onnx.save(model, model_path)

        # Declare MinMax calibrator from model path
        calibrator = create_calibrator(
            model_path,
            op_types_to_calibrate=_get_op_types_to_calibrate(),
            augmented_model_path=tmp_dir.joinpath("augmented_model.onnx"),
            calibrate_method=CalibrationMethod.MinMax,
            use_external_data_format=False,
            extra_options={"symmetric": symmetric,
                           "moving_average": average,
                           "per_channel": not per_tensor_activations})

        # Collect output tensors with calibration data and compute range
        calibrator.collect_data(calibration_data_reader)
        tensors_range = calibrator.compute_data()
        del calibrator
    return tensors_range



__all__ = ["load_model", "save_model"]

import os
import keras
import onnx


def load_model(model_path, custom_layers=None, compile_model=True):
    """
    Loads an Onnx or Keras model. An error is raised if the provided model extension is not
    supported.

    Args:
        model_path (str): path of the model to load.
        custom_layers (dict, optional): custom layers to add to the Keras model. Defaults to None.
        compile_model (bool, optional): whether to compile the Keras model. Defaults to True.

    Returns:
        keras.models.Model or onnx.ModelProto: Loaded model.

    Raises:
        ValueError: if the model could not be loaded using Keras and ONNX loaders.
    """
    _, model_extension = os.path.splitext(model_path.lower())

    if model_extension == '.h5':
        model = keras.models.load_model(model_path,
                                        custom_objects=custom_layers,
                                        compile=compile_model)
    elif model_extension == '.onnx':
        model = onnx.load_model(model_path)
    else:
        raise ValueError(
            f"Unsupported model extension: '{model_extension}'. "
            f"Expected model with extension(s): {['h5', 'onnx']}"
        )

    return model


def save_model(model, path):
    """
    Save an ONNX or Keras model into a path.

    Note extension is overwritten given the model type.

    Args:
        model (keras.Model, keras.Sequential or onnx.ModelProto): model to serialize.
        model_path (str): path to save the model.

    Returns:
        str: the path where the model was saved.

    Raises:
        ValueError: if the model to save is not a Keras or ONNX model.
    """
    model_name, _ = os.path.splitext(path)
    if isinstance(model, (keras.Model, keras.Sequential)):
        model_path_with_ext = model_name + ".h5"
        model.save(model_path_with_ext, include_optimizer=False)
    elif isinstance(model, onnx.ModelProto):
        model_path_with_ext = model_name + ".onnx"
        onnx.save_model(model, model_path_with_ext)
    else:
        raise ValueError(f"Unrecognized {type(model)} model type. Expected a keras or ONNX model.")
    return model_path_with_ext



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["Add", "QuantizedAdd", "QuantizedConcatenate"]

import keras
import tensorflow as tf

from .layers_base import (register_quantize_target, register_no_output_quantizer,
                          register_aligned_inputs, apply_buffer_bitwidth, QuantizedLayer)
from .quantizers import OutputQuantizer
from .recorders import TensorRecorder
from ..tensors import FixedPoint


@keras.saving.register_keras_serializable()
class Add(keras.layers.Layer):
    """Wrapper class of `keras.layers.Add` that allows to average inputs.

    We only support a tuple of two inputs with the same shape.

    Args:
        average (bool, optional): if `True`, compute the average across all inputs.
            Defaults to `False`.
        activation (bool, optional): If `True`, apply an activation function after
            the addition. Defaults to `False`.
    """

    def __init__(self, *args, average=False, activation=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.average = average
        self.activation = activation

    def build(self, input_shape):
        if not isinstance(input_shape, (list, tuple)) or len(input_shape) != 2:
            raise ValueError(f"{self.__class__.__name__} expects two input tensors")
        super().build(input_shape)

    def call(self, inputs):
        a, b = inputs
        output = tf.add(a, b)
        if self.average:
            output /= 2
        if self.activation:
            output = tf.nn.relu(output)
        return output

    def get_config(self):
        config = super().get_config()
        config["average"] = self.average
        config["activation"] = self.activation
        return config


# Note that keras.layers.Add is not registered as a quantized target because the class name 'Add' is
# enough and shared between Keras and QuantizeML
@register_quantize_target(Add)
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedAdd(QuantizedLayer, Add):
    """Sums two inputs and quantize the output.

    The two inputs must be provided as a list or tuple of FixedPoint or Tensors.

    The outputs are quantized according to the specified quantization configuration.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """

    def __init__(self, *args, quant_config=None, **kwargs):
        super().__init__(*args, quant_config=quant_config, **kwargs)
        self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=True)
        out_quant_cfg = self.quant_config.get("output_quantizer", False)
        if out_quant_cfg:
            self.out_quantizer = OutputQuantizer(
                name="output_quantizer", **out_quant_cfg)
        else:
            self.out_quantizer = None
        # Add objects that will store the shift values.
        self.a_shift = TensorRecorder(self.name + "/a_shift")
        self.b_shift = TensorRecorder(self.name + "/b_shift")

    def _quantized_relu(self, inputs):
        # Inputs are always fixed points (as they come after an Add or after an output quantizer)
        # we can then represent the zero as a fixed point
        zero = FixedPoint(tf.constant(0.), inputs.value_bits, inputs.frac_bits)
        # ReLU inside the Add is always unbounded so we just remove negative values
        return tf.math.maximum(inputs, zero)

    def call(self, inputs):
        a, b = inputs
        if not (isinstance(a, FixedPoint) and isinstance(b, FixedPoint)):
            # If any of the inputs is not a FixedPoint, raise an error
            raise TypeError(f"QuantizedAdd only accepts FixedPoint\
                              inputs. Receives {(type(a), type(b))} inputs.")

        # Align intermediate inputs before adding them
        a, shift_ab = a.align(b, self.buffer_bitwidth)
        b, shift_ba = b.align(a, self.buffer_bitwidth)

        outputs = tf.add(a, b)
        if self.average:
            outputs = outputs >> 1

        # Compute shifts
        self.a_shift(shift_ab)
        self.b_shift(shift_ba)

        # Rescale outputs
        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs)

        # Apply activation
        if self.activation:
            outputs = self._quantized_relu(outputs)

        return outputs


@register_quantize_target(keras.layers.Concatenate)
@register_no_output_quantizer
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedConcatenate(QuantizedLayer, keras.layers.Concatenate):
    """ A Concatenate layer that operates on quantized inputs
    """

    def call(self, inputs):
        a, b = inputs
        if not (isinstance(a, FixedPoint) and isinstance(b, FixedPoint)):
            # If any of the inputs is not a FixedPoint, raise an error
            raise TypeError(f"QuantizedConcatenate only accepts FixedPoint\
                              inputs. Receives {(type(a), type(b))} inputs.")

        return tf.concat([a, b], axis=self.axis)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["shiftmax", "Shiftmax", "QuantizedShiftmax"]

import tensorflow as tf
import keras

from .layers_base import (register_quantize_target, register_no_output_quantizer,
                          register_aligned_inputs, apply_buffer_bitwidth, QuantizedLayer)
from .recorders import TensorRecorder
from ..tensors import FixedPoint, floor_through, round_log2, pow2


def shiftmax(logits, axis=-1):
    """Computes softmax-like activations, but using base 2 for the exponential.

    Used as approximation of the softmax activation.

    This function performs the equivalent of

        >>>  logits = tf.floor(logits)
        >>>  exp = 2 ** logits
        >>>  sum_exp_shift = tf.round(tf.log2(tf.reduce_sum(exp, axis, keepdims=True)))
        >>>  softmax = exp / 2 ** sum_exp_shift = 2 ** (logits -  sum_exp_shift)


    When 2 ** :attr:`sum_exp_shift` is an approximated of sum_exp as a Power-of-Two (PoT)

    To avoid a high exponential (and a tf.inf representation by tensorflow), we adopt the
    following equivalence:

    Making the variable change :math:`y=logits-x0`, we reach the same result as
    :math:`p=shiftmax(logits)`, because,

    .. math::
        p' = \\frac{2^y}{sum(2^y)}
           = \\frac{2^{logits-x0}}{sum(2^{logits-x0})}
           = \\frac{2^{logits} * 2^{-x0}}{2^{-x0} * sum(2^{logits})}
           = \\frac{2^{logits}}{sum(2^{logits})}
           = p

    We take :math:`x0 = max(logits)`.

    Args:
        logits (tf.Tensor): a non-empty `Tensor`.
        axis (int, list, optional): the dimension shiftmax would be performed
            on. The default is -1 which indicates the last dimension.

    Returns:
        tf.Tensor: value of shiftmax function with the same type and shape as `logits`.

    Raises:
        InvalidArgumentError: if `logits` is empty or `axis` is beyond the last
            dimension of `logits`.

    Note:
        We floor the :attr:`logits` to approximate the results to those expected
        when quantizing the operation.
    """
    logits = floor_through(logits)
    logits = logits - tf.reduce_max(logits, axis=axis, keepdims=True)
    exp = tf.cast(2**logits, dtype=logits.dtype)
    sum_exp = tf.reduce_sum(exp, axis=axis, keepdims=True)
    sum_exp_shift = round_log2(sum_exp)
    return 2 ** (logits - sum_exp_shift)


@keras.saving.register_keras_serializable()
class Shiftmax(keras.layers.Layer):
    """Wrapper class of `shiftmax` function, that calculates a softmax-like
    activation.

    Note that shiftmax operation is performed always along the last axis.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def call(self, inputs):
        output = shiftmax(inputs)
        return output


@register_quantize_target(Shiftmax)
@register_no_output_quantizer
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedShiftmax(QuantizedLayer):
    """A quantized layer to do a quantized function similar to the softmax, but
    using base 2 instead of e. So we replace

    .. math:: softmax(x_i) = \\frac{e^{x_i}}{sum(e^{x_k})}

    With this:

    .. math:: softmax2(x_i) = \\frac{2^{x_i}}{sum(2^{x_k})}

    This approximation is close enough to the original function. In order to
    make it more hardware friendly, we also approximated the :math:`sum(2^{x_k})`
    to the closest power of two:

    .. math:: shiftmax(x_i) = \\frac{2^{x_i}}{2^{round(log2(sum(2^{x_k})))}}

    So it can be implemented with a simple shift operation.

    Implementation is inspired from this paper:

    Cardarilli, G.C., Di Nunzio, L., Fazzolari, R. et al.
    A pseudo-softmax function for hardware-based high speed image classification.
    Sci Rep 11, 15307 (2021). https://doi.org/10.1038/s41598-021-94691-7

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """

    def __init__(self, *args, quant_config=None, **kwargs):
        super().__init__(*args, quant_config=quant_config, **kwargs)
        self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=True)
        self.exp_bitwidth = self.quant_config.get("exp_bitwidth", 10)
        if self.buffer_bitwidth <= 2 * self.exp_bitwidth:
            raise ValueError(f"exp_bitwidth={self.exp_bitwidth} must be less than "
                             f"half of buffer_size={self.buffer_bitwidth}.")
        # Add objects that will store the shift values.
        self.input_shift = TensorRecorder(self.name + "/input_shift")

    def call(self, x):
        # Raise an error if the inputs are not FixedPoint
        if not isinstance(x, FixedPoint):
            raise TypeError("QuantizedShiftmax only accepts FixedPoint inputs. Receives "
                            f"{type(x)} inputs.")

        # To avoid overflowing, some modifications are made to the input.
        # First remove the fractional part of the input (floor(x)). We can do
        # this because the exponential function will return very big numbers,
        # so fractional ones can be ignored in the ratio with the sum.
        x, shift = x.floor()
        # update shift values if recording is enabled
        self.input_shift(shift)

        # Since x has been floored, we can directly use its values
        x = x.values

        # The pseudo-softmax is defined as:
        #
        # p = 2^x/sum(2^x)
        #
        # but we do this instead:
        #
        # p' = p = 2^y/sum(2^y)
        #
        # where
        #
        # y = x - x0
        #
        # because,
        #
        # p' = 2^y/sum(2^y) = 2^(x - x0)/sum(2^(x - x0)) = (2^x * 2^-x0)/(2^-x0 * sum(2^x))
        #    = 2^x/sum(2^x) = p
        #
        # On the other hand, we choose x0 to be the maximum of x, minus a positive
        # integer constant "exp_bitwidth". So now
        #
        # y = x - (max(x) - exp_bitwidth)
        #
        # This makes sure that y is never higher than exp_bitwidth
        x_max = tf.reduce_max(x, axis=-1, keepdims=True)
        x0 = x_max - self.exp_bitwidth
        y = x - x0

        # To evaluate exp = 2^y, we target a maximum precision of exp_bitwidth.
        # As a consequence, we will neglect all values that are below -exp_bitwidth,
        # considering:
        # - that they don't contribute much to the exponential sum,
        # - that they would be quantized to zero after the division.
        exp_values = tf.where(y >= -self.exp_bitwidth, pow2(y + self.exp_bitwidth), 0)
        # Note that we could do the operation directly on the values, but we store
        # values in a FixedPoint to make sure we don't saturate the underlying buffer
        exp = FixedPoint(exp_values, self.buffer_bitwidth, self.exp_bitwidth)
        # To calculate 2^y, hardware can just:
        # - set exp to zero if y < -exp_bitwidth,
        # - do a left shift applying a fixed offset of self.exp_bitwidth.
        # Example:
        #   exp_bitwidth = 4
        #   y = [-5, 3, -4, -1, 1]
        #   exp = [0, 1 << (4 + 3), 1 << (4 - 4), 1 << (4 - 1), 1 << (4 + 1)]
        #   exp = [0, 128, 1, 8, 32]

        # Calculate the sum of the exponential (saturation may happen here).
        sum_exp = tf.reduce_sum(exp, axis=-1, keepdims=True)

        # Like the float version, instead of dividing by sum_exp, we simply approximate
        # it to the closest integer log2 to perform a shift instead of a division.
        # Please refer to the description of round_log2 for a description of the hardware operation.
        # Note here that we need to substract the frac_bits as the values are scaled up.
        sum_exp_shift = round_log2(sum_exp.values) - sum_exp.frac_bits
        outputs = exp.shift(-sum_exp_shift)

        # Since sum_exp > exp, the results are between [0,1].
        # We can therefore rewrite the output as:
        return FixedPoint(outputs.values, self.exp_bitwidth + 1, self.exp_bitwidth)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from ...debugging import assert_equal
from ..fixed_point import FixedPoint
from ..qfloat import QFloat


@tf.experimental.dispatch_for_api(tf.add)
def fp_add(x: FixedPoint, y: FixedPoint):
    return x + y


@tf.experimental.dispatch_for_api(tf.add)
def qf_add(x: QFloat, y: QFloat):
    return x + y


@tf.experimental.dispatch_for_api(tf.clip_by_value)
def fp_clip_by_value(t: FixedPoint,
                     clip_value_min: FixedPoint,
                     clip_value_max: FixedPoint,
                     name=None):
    """Clips tensor values to a specified min and max.

    Args:
        t (:obj:`FixedPoint`): the FixedPoint to be clipped.
        clip_value_min (:obj:`FixedPoint`): the minimum value.
        clip_value_max (:obj:`FixedPoint`): the maximum value.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the clipped FixedPoint.
    """
    # The fractional bits of all inputs must be equal
    assert_equal(t.frac_bits, clip_value_min.frac_bits)
    assert_equal(t.frac_bits, clip_value_max.frac_bits)
    clip_values = tf.clip_by_value(t.values, clip_value_min.values, clip_value_max.values, name)
    return FixedPoint(clip_values, t.value_bits, t.frac_bits)


@tf.experimental.dispatch_for_api(tf.clip_by_value)
def qf_clip_by_value(t: QFloat,
                     clip_value_min: QFloat,
                     clip_value_max: QFloat,
                     name=None):
    """Clips tensor values to a specified min and max.

    Args:
        t (:obj:`QFloat`): the QFloat to be clipped.
        clip_value_min (:obj:`QFloat`): the minimum value.
        clip_value_max (:obj:`QFloat`): the maximum value.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`QFloat`: the clipped QFloat.
    """
    # The scales of all inputs must be equal
    assert_equal(t.scales, clip_value_min.scales)
    assert_equal(t.scales, clip_value_max.scales)
    clip_fp = tf.clip_by_value(t.fp, clip_value_min.fp, clip_value_max.fp, name)
    return QFloat(clip_fp, t.scales)


@tf.experimental.dispatch_for_api(tf.math.maximum)
def fp_maximum(x: FixedPoint, y: FixedPoint, name=None):
    """Returns the max of x and y element-wise.

    Args:
        x (:obj:`FixedPoint`): the first FixedPoint.
        y (:obj:`FixedPoint`): the second FixedPoint.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the resulting FixedPoint.
    """
    # The inputs frac_bits must be equal
    assert_equal(x.frac_bits, y.frac_bits)
    # Choose the highest of both value bits
    value_bits = max(x.value_bits, y.value_bits)
    return FixedPoint(tf.math.maximum(x.values, y.values, name), value_bits, x.frac_bits)


@tf.experimental.dispatch_for_api(tf.math.maximum)
def qf_maximum(x: QFloat, y: QFloat, name=None):
    """Returns the max of x and y element-wise.

    Args:
        x (:obj:`QFloat`): the first QFloat.
        y (:obj:`QFloat`): the second QFloat.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`QFloat`: the resulting QFloat.
    """
    # The inputs scales must be equal
    assert_equal(x.scales, y.scales)
    max_fp = tf.math.maximum(x.fp, y.fp, name)
    return QFloat(max_fp, x.scales)


def _check_last_dim_reduced(input_tensor, axis):
    last_dim = input_tensor.shape.ndims - 1
    last_dim_reduced = axis is None or axis == -1 or axis == last_dim
    if isinstance(axis, (list, tuple)):
        last_dim_reduced = last_dim_reduced or last_dim in axis
    if last_dim_reduced:
        # The last dimension will be collapsed: inputs must be quantized per-tensor
        input_tensor.assert_per_tensor()


@tf.experimental.dispatch_for_api(tf.math.reduce_sum)
def fp_reduce_sum(input_tensor: FixedPoint, axis=None, keepdims=False, name=None):
    """Computes the sum of elements across dimensions of a FixedPoint.

    Args:
        input_tensor (:obj:`FixedPoint`): the FixedPoint to be summed.
        axis (list, optional): the dimensions to reduce. If None, reduces all
            dimensions. Defaults to None.
        keepdims (bool, optional): if true, retains reduced dimensions with length 1.
            Defaults to False.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the summed FixedPoint.
    """
    _check_last_dim_reduced(input_tensor, axis)

    # Reduce sum
    s_values = tf.math.reduce_sum(
        input_tensor.values, axis, keepdims=keepdims, name=name)

    # Return a new FixedPoint
    return FixedPoint(s_values, input_tensor.value_bits, input_tensor.frac_bits)


@tf.experimental.dispatch_for_api(tf.math.reduce_sum)
def qf_reduce_sum(input_tensor: QFloat, axis=None, keepdims=False, name=None):
    """Computes the sum of elements across dimensions of a QFloat.

    Args:
        input_tensor (:obj:`QFloat`): the QFloat to be summed.
        axis (list, optional): the dimensions to reduce. If None, reduces all
            dimensions. Defaults to None.
        keepdims (bool, optional): if true, retains reduced dimensions with length 1.
            Defaults to False.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`QFloat`: the summed QFloat.
    """
    _check_last_dim_reduced(input_tensor, axis)

    # Reduce sum
    s_fp = tf.math.reduce_sum(input_tensor.fp, axis, keepdims=keepdims, name=name)

    # Return a new QFloat with the same scales
    return QFloat(s_fp, input_tensor.scales)


@tf.experimental.dispatch_for_api(tf.math.reduce_max)
def fp_reduce_max(input_tensor: FixedPoint,
                  axis=None,
                  keepdims=False,
                  name=None):
    """Computes the maximum of elements across dimensions of a FixedPoint.

    Args:
        input_tensor (:obj:`FixedPoint`): the FixedPoint to be estimated.
        axis (list, optional): the dimensions to reduce. If None, reduces all
            dimensions. Defaults to None.
        keepdims (bool, optional): if true, retains reduced dimensions with length 1.
            Defaults to False.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the maximum FixedPoint.
    """
    # We only support reduce_max if the input is per-tensor
    input_tensor.assert_per_tensor()

    # Reduce max
    s_values = tf.math.reduce_max(input_tensor.values,
                                  axis,
                                  keepdims=keepdims,
                                  name=name)
    # Return a new FixedPoint
    return FixedPoint(s_values, input_tensor.value_bits, input_tensor.frac_bits)


@tf.experimental.dispatch_for_api(tf.math.multiply)
def fp_fp_multiply(x: FixedPoint, y: FixedPoint, name=None):
    """Returns an element-wise x * y.

    Args:
        x (:obj:`FixedPoint`): a FixedPoint
        y (:obj:`FixedPoint`): a FixedPoint
        name (str, optional): A name for the operation

    Returns:
        :obj:`FixedPoint`: an object with the same shape as x
    """
    values = tf.math.multiply(x.values, y.values, name)
    return FixedPoint(values, x.value_bits, x.frac_bits + y.frac_bits)


@tf.experimental.dispatch_for_api(tf.math.multiply)
def fp_qf_multiply(x: FixedPoint, y: QFloat, name=None):
    """Returns an element-wise x * y.

    Args:
        x (:obj:`FixedPoint`): a FixedPoint
        y (:obj:`QFloat`): a QFloat
        name (str, optional): A name for the operation

    Returns:
        :obj:`QFloat`: an object with the same shape as x
    """
    # Multiply x by y inner FixedPoint
    # To produce a per-tensor FixedPoint, x must be per-tensor
    x.assert_per_tensor()
    mul = tf.math.multiply(x, y.fp, name)
    # Return a new QFloat with y scales
    return QFloat(mul, y.scales)


@tf.experimental.dispatch_for_api(tf.math.multiply)
def qf_qf_multiply(x: QFloat, y: QFloat, name=None):
    """Returns an element-wise x * y.

    Args:
        x (:obj:`QFloat`): a QFloat
        y (:obj:`QFloat`): a QFloat
        name (str, optional): A name for the operation

    Returns:
        :obj:`QFloat`: an object with the same shape as x
    """
    # To produce a per-tensor FixedPoint, x must be per-tensor
    x.assert_per_tensor()
    # Multiply x and y inner FixedPoints
    mul = tf.math.multiply(x.fp, y.fp, name)
    # Evaluate the resulting scales as the product of both scales
    scales = x.scales * y.scales
    # Return a new QFloat
    return QFloat(mul, scales)


@tf.experimental.dispatch_for_api(tf.math.divide)
def fp_fp_divide(x: FixedPoint, y: FixedPoint, name=None):
    """Computes Python style division of `x` by `y`.

    Args:
        x (:obj:`FixedPoint`): the FixedPoint to divide
        y (:obj:`FixedPoint`): the FixedPoint divisor
        name (str, optional): A name for the operation

    Returns:
        :obj:`FixedPoint`: an object with the same shape as x
    """
    return x / y


@tf.experimental.dispatch_for_api(tf.math.divide)
def qf_fp_divide(x: QFloat, y: FixedPoint, name=None):
    """Computes Python style division of `x` by `y`.

    Args:
        x (:obj:`QFloat`): the QFloat to divide
        y (:obj:`FixedPoint`): the FixedPoint divisor
        name (str, optional): A name for the operation

    Returns:
        :obj:`QFloat`: an object with the same shape as x
    """
    # Divide x inner FixedPoint by y
    # To produce a per-tensor FixedPoint, y must be per-tensor
    y.assert_per_tensor()
    div = tf.math.divide(x.fp, y, name)
    # Return a new QFloat with x scales
    return QFloat(div, x.scales)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Transformation to split a layer with ReLu activation into two layers:
A first layer with the weigths and a second layer with only ReLu.
"""


__all__ = ["split_layers_with_relu"]

from .insert_layer import insert_in_config
from .transforms_utils import get_layer_index
from ..utils import apply_weights_to_model

from keras.layers import ReLU


def split_layers_with_relu(model):
    """
    Splits layers with 'relu' activation into two separate layers:
    one without activation and one with a separate 'relu' activation layer.

    Args:
        model (keras.Model): The Keras model to be modified (sequential or functional).

    Returns:
        Model: The updated Keras model with 'relu' activations split into separate layers.
    """

    config = model.get_config()
    for layer in model.layers:
        if ('activation' in dir(layer) and
                (not isinstance(layer.activation, bool)) and
                layer.activation.__name__ == 'relu'):
            # Modify the first layer
            index = get_layer_index(config['layers'], layer.name)
            config['layers'][index]['config']['activation'] = 'linear'
            # Create the new layer
            config_no_activation = layer.get_config()
            activation_layer_name = config_no_activation['name'] + '_activ'
            activation_layer = ReLU(name=activation_layer_name)
            # Insert the second layer (only ReLu) into the configuration
            insert_in_config(model, layer.name, activation_layer, config)
    # update the model
    new_model = model.from_config(config)
    # Load original weights
    variables_dict = {var.name: var for var in model.variables}
    apply_weights_to_model(new_model, variables_dict, False)
    return new_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from .qtensor import QTensor, slice_tensor_by_index
from .fixed_point import FixedPoint
from ..debugging import assert_equal


class QFloat(QTensor):
    """A Tensor of FixedPoint values and scales representing float numbers

    The QFloat is a dual representation of a float Tensor combining FixedPoint
    values and float scales.

    The QFloat is typically used to represent float tensors whose quantization
    range is not 'optimal' for FixedPoint quantization: the original tensor is
    first divided by the scales to be aligned on optimal ranges, then quantized
    to FixedPoint values.

    When converting back to float, values are dequantized and multiplied by the
    scales to obtain the approximated float tensor.

    Args:
        fp (:obj:`FixedPoint`): a FixedPoint of values
        scales (tf.Tensor): a Tensor of scales
    """
    fp: FixedPoint = FixedPoint(1.0, 32, 0)
    scales: tf.Tensor = 1.0

    def __init__(self, fp, scales):
        self.scales = tf.cast(scales, tf.float32)
        self.fp = fp
        self.shape = self.fp.shape

    @property
    def values(self):
        return self.fp.values

    @property
    def frac_bits(self):
        return self.fp.frac_bits

    @property
    def value_bits(self):
        return self.fp.value_bits

    @property
    def name(self):
        return self.fp.name

    @property
    def per_tensor(self):
        return self.fp.per_tensor and self.scales.shape.ndims == 0

    @staticmethod
    def max_frac_bits(value_bits, ranges, scales, clamp=True):
        """Evaluate the maximum fractional bit index for the quantization ranges.

        This method evaluates the minimum number of integer bits required to cover
        the specified quantization ranges after having rescaled them with the specified
        scales.
        It simply calls the equivalent FixedPoint method on the rescaled ranges.
        If specified, it clamps the results to the available value_bits.

        Args:
            value_bits (int): the number of value bits.
            ranges (tf.Tensor): a tensor of float quantization ranges.
            scales (tf.Tensor): the scales to apply to the quantization ranges.
            clamp (bool, optional): clamp the results to self.value_bits. Defaults to True.

        Returns:
            tf.Tensor: a tensor of fractional bits.
        """
        return FixedPoint.max_frac_bits(value_bits, ranges / scales, clamp)

    @staticmethod
    def optimal_scales(ranges, value_bits):
        r"""Evaluates the optimal QFloat scales for quantization ranges.

        We choose the optimal quantization range for a given bitwidth as:

            [-int_max, int_max], with :math:`int\_max = 2^{bits} - 1`.

        This methods evaluates the scales as the ratio to align the specified ranges
        to the optimal ranges.

        Args:
            ranges (tf.Tensor): a tensor of quantization ranges.
            value_bits (int): the number of value bits.

        Returns:
            tf.Tensor: the optimal scales.
        """
        # Evaluate the scales as the ratio to align the ranges on the optimal quantization range
        optimal_ranges = QTensor.int_max(value_bits)
        scales = tf.cast(ranges, tf.float32) / optimal_ranges
        return scales

    @staticmethod
    def quantize_scales(scales, scale_bits):
        """Quantizes the QFloat scales with the specified bitwidth.

        Args:
            scales (tf.Tensor): a tensor of float scales.
            scale_bits (int): the number of scales bits.

        Returns:
            :obj:`FixedPoint`: the FixedPoint scales.
        """
        # Evaluate QFloat scales fractional bits, without clamping to scale_bits
        scales_frac_bits = FixedPoint.max_frac_bits(scale_bits, scales, clamp=False)
        return FixedPoint.quantize(scales, scale_bits, scales_frac_bits)

    @staticmethod
    def quantize(x, value_bits, scales, frac_bits=0.):
        """Converts a float Tensor to a QFloat

        It first evaluates and quantizes the scales required to align the quantization ranges
        to the optimal range for the specified value bits.

        It then quantizes the inputs with the quantized scales.

        The resulting integer values are clipped to [-int_max-1, int_max].

        Args:
            x (tf.Tensor): a tensor of float values.
            value_bits (int): the number of value bits.
            scales (tf.Tensor): a tensor of alignment scales.
            frac_bits (int): the inner FixedPoint fractional bits (defaults to 0).

        Returns:
            :obj:`QFloat`: the QFloat representation.
        """
        # Rescale the inputs to project them in the actual quantization ranges
        x_scaled = x / scales
        # Quantize the rescaled inputs with the specified fractional bits
        fp_x_scaled = FixedPoint.quantize(x_scaled, value_bits, frac_bits)
        return QFloat(fp_x_scaled, scales)

    def to_float(self):
        """Returns a float representation of the QFloat

        Returns:
            tf.Tensor: the float representation.
        """
        return self.fp.to_float() * self.scales

    def to_fixed_point(self, scale_bits=8):
        """Returns a FixedPoint representation of the QFloat

        Args:
            scale_bits (int, optional): the scales quantization bitwidth. Defaults to 8.

        Returns:
            (:obj:`FixedPoint`, :obj:`FixedPoint`): the FixedPoint representation and scales.
        """
        # Quantize the scales
        scales = QFloat.quantize_scales(self.scales, scale_bits)
        # Return the FixedPoint product of scales and values
        return tf.math.multiply(self.fp, scales, name="apply_scales"), scales

    def upscale(self, frac_bits, value_bits=None):
        """Align a QFloat to a specified precision

        The target precision must be higher than the current one.

        Args:
            frac_bits (tf.Tensor): the target fractional bits
            value_bits (int, optional): the target value bits (defaults to current value bits)

        Returns:
            :obj:`FixedPoint`: the upscaled FixedPoint
        """
        if value_bits is None:
            value_bits = self.value_bits
        # Upscale the inner values FixedPoint
        fp, shift = self.fp.upscale(frac_bits, value_bits)
        # Return a new QFloat with updated values and the same scale
        return QFloat(fp, self.scales), shift

    def expand(self, value_bits):
        """Expand the QFloat to the specified bitwidth

        This returns an equivalent QFloat with a higher or equal number of
        value bits and a scalar fractional bit corresponding to the maximum of
        the initial fractional bits on all channels. The scales remains unchanged.

        This is mostly used to recover a per-tensor QFloat that has been
        compressed to a lower number of value bits.

        Note that even if the frac_bits are aligned the scales remained unchanged.

        Args:
            value_bits (int): the target value_bits

        Returns:
            tuple(:obj:`QFloat`, tf.Tensor): a new QFloat with expanded
            fractional bits and the shift that was applied.
        """
        if value_bits < self.value_bits:
            raise ValueError(
                f"Cannot reduce {self.name} bitwidth from {self.value_bits} to {value_bits}:"
                " use a quantizer instead.")
        max_frac_bits = tf.reduce_max(self.frac_bits)
        return self.upscale(max_frac_bits, value_bits)

    def promote(self, bits):
        """Increase the number of value bits

        Args:
            bits (int): the new number of value bits

        Returns:
            :obj:`QFloat`: a QFloat with increased value bits
        """
        # Return a new QFloat with identical scales and a promoted inner FixedPoint
        return QFloat(self.fp.promote(bits), self.scales)

    def __add__(self, other):
        if isinstance(other, QFloat):
            # Check that self and other have the same scales
            assert_equal(self.scales, other.scales,
                         message=f"{self.name} and {other.name} have different scales.")
            # Return a new QFloat
            return QFloat(self.fp + other.fp, self.scales)
        raise TypeError(
            f"Unsupported operand type(s) for +: 'QFloat' and '{type(other)}'")

    def __sub__(self, other):
        if isinstance(other, QFloat):
            # Check that self and other have the same scales
            assert_equal(self.scales, other.scales,
                         message=f"{self.name} and {other.name} have different scales.")
            # Return a new QFloat
            return QFloat(self.fp - other.fp, self.scales)
        raise TypeError(
            f"Unsupported operand type(s) for -: 'QFloat' and '{type(other)}'")

    def __truediv__(self, other):
        # We only support the division of the scales by a compatible type
        div_scales = self.scales / other
        # Return a Qfloat with updated scales
        return QFloat(self.fp, div_scales)

    def __getitem__(self, idx):

        r"""Retrieve a slice or element from the QFloat tensor.

        This function allows for slicing or indexing the QFloat tensor in a similar way
        to a standard TensorFlow tensor. It handles the slicing of both the fixedpoint
        and the scales of the QFloat tensor, ensuring that the tensor
        representation remains consistent.

        Args:
            idx (int, slice, or tuple): The index or slice to retrieve. It can be an integer,
                a slice, or a tuple of integers/slices. The tuple can
                also include the Ellipsis (`...`) to denote a slice
                over remaining dimensions.

        Returns:
            QFloat: A new QFloat instance containing the sliced values and corresponding
                scales.
        """

        sliced_fp = self.fp[idx]

        if self.per_tensor:
            return QFloat(sliced_fp, self.scales)

        else:
            values_rank = len(self.shape)
            sliced_scales = slice_tensor_by_index(idx, values_rank, self.scales)
            return QFloat(sliced_fp, sliced_scales)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["plot_kernel_distribution"]

import os
import matplotlib.pyplot as plt

import onnx
from tensorboardX import SummaryWriter
from keras.layers import DepthwiseConv2D

from ..models import record_quantization_variables
from ..layers.layers_base import QuantizedLayer, WeightQuantizer
from ..models.transforms import sanitize as keras_sanitize
from ..onnx_support.quantization import ONNXModel
from ..onnx_support.quantization.transforms import sanitize as onnx_sanitize


def _plot_kernel_distribution_on_writer(writer, layer_name, kernel):
    kernel = kernel.reshape((kernel.shape[0], -1) if kernel.ndim > 1 else (1, -1))
    for idx, k in enumerate(kernel):
        writer.add_histogram(f"histograms/{layer_name}", k, idx)

    # Boxplot
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(16, 9))
    ax.boxplot(kernel.transpose(),
               patch_artist=True,
               vert=True,
               notch=True,
               boxprops={"facecolor": "orange"},
               medianprops={"color": "gray", "linewidth": 1.5},
               labels=[f"k{x}" for x in range(kernel.shape[0])])

    # Increase figure dpi and reduce pad layout
    fig.set_dpi(300)
    fig.tight_layout()
    writer.add_figure(f"boxplots/{layer_name}", figure=fig)


def _plot_onnx_kernel_distribution(model, logdir):
    def _get_kernel_name(node):
        skip_qnodes = ("QuantizedAdd", "InputQuantizer", "Dequantizer")
        if node.domain == "com.brainchip" and not any(sn in node.op_type for sn in skip_qnodes):
            # kernel in InputQuantizedConv is moved to position 2
            return node.input[1] if "InputConv" not in node.op_type else node.input[2]
        elif node.op_type in ("Conv", "Gemm"):
            return node.input[1]
        return None

    model = onnx_sanitize(model)
    logdir = os.path.join(logdir, model.name or "model")
    with SummaryWriter(log_dir=logdir) as writer:
        for idx, node in enumerate(model.nodes()):
            if (kernel_name := _get_kernel_name(node)):
                kernel = model.get_variable(kernel_name)
                # Unlike Keras, ONNX does not require nodes to be named.
                # In that case, we group the graphs by their id
                _plot_kernel_distribution_on_writer(writer, node.name or f"node_{idx}", kernel)


def _plot_keras_kernel_distribution(model, logdir):
    model = keras_sanitize(model)
    record_quantization_variables(model)
    logdir = os.path.join(logdir, model.name or "model")
    with SummaryWriter(log_dir=logdir) as writer:
        for layer in model.layers:
            weights = layer.get_weights()
            if len(weights) > 0:
                if isinstance(layer, QuantizedLayer):
                    # Search WeightQuantizer in layer
                    weights = None
                    for attr in layer.__dict__.values():
                        if isinstance(attr, WeightQuantizer):
                            # Replace weights by its quantized version
                            weights = [attr.qweights.value.values.numpy()]
                            break
                    # Skip quantized layer if it does not have a WeightQuantizer
                    if weights is None:
                        continue
                weights = weights[0]
                if weights.ndim >= 2:
                    # Weights are formartted as ONNX
                    list_axis = list(range(weights.ndim - 2))
                    if not isinstance(layer, DepthwiseConv2D):
                        weights = weights.transpose((-1, -2, *list_axis))
                    else:
                        # Axis (0,1) should be exchanged if layer is a DepthwiseConv2D
                        weights = weights.transpose((-2, -1, *list_axis))
                _plot_kernel_distribution_on_writer(writer, layer.name, weights)


def plot_kernel_distribution(model, logdir):
    """Plot the kernel distribution of each layer/node in the model.

    Distributions are plotted in two ways: histogram and boxplot

    After exporting them, the plots can be plotted through the command-line:

    >>> tensorboard --logdir=`logdir`

    Args:
        model (onnx.ModelProto or tf.keras.Model): the model to plot the kernel distribution
        logdir (str): the directory to save the plots
    """
    if isinstance(model, onnx.ModelProto):
        _plot_onnx_kernel_distribution(ONNXModel(model), logdir)
    else:
        _plot_keras_kernel_distribution(model, logdir)
    print(f"[INFO] Plots were saved on '{logdir}' successfully. They are available through "
          f"the command-line:\n>>> tensorboard --logdir={logdir}")



from .base_layer import *
from .quantizers import *
from .conv2d import *
from .conv2dtranspose import *
from .depthwise2d import *
from .dense import *
from .multi_inbounds import *
from .register import *
from .layer_compatibility import *



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_padding_ops"]

import numpy as np

from onnx.helper import make_node


def transform_pads_into_array(pads):
    """Return the expected pads to apply in the custom operation.

    Args:
        pads (list of int): the pads to transform.

    Returns:
        TensorProto: pads as tensor.
    """
    assert len(pads) == 4, "Expect 4 values for pads"
    # ONNX Pad expect a 1D tensor of shape [2 * num_axes]. Given we should apply pads
    # over XY dimensions, others will be set to 0.
    # See https://onnx.ai/onnx/operators/onnx__Pad.html for more information.
    pads = [0, 0] + pads[:2] + [0, 0] + pads[2:]
    return np.array(pads, "int64")


def get_padding_ops(in_name, out_name, pad_value_name=""):
    """Return the pad operation chain.

    Args:
        in_name (str): the input tensor name.
        out_name (str): the required output tensor name.
        pad_value_name (str, optional): name of padding value.
            Takes a zero value when not specified. Defaults to "".

    Returns:
        list of NodeProto: the operation chain.
    """
    if not pad_value_name:
        return [make_node("Pad", inputs=[in_name, "pads"], outputs=[out_name])]

    # Knowledge: padding_value is a tensor and it contains one value per input channel.
    # The following sequence of operations is carried out to apply padding_value for each channel
    # X -> Sub(X, padding_value) -> Pad(Xsub, pad) -> Add(Xpadded, padding_value) -> Y

    nodes = [make_node("Constant", [], ["pad_axes"], value_ints=[1, 2]),
             make_node("Unsqueeze", [pad_value_name, "pad_axes"], ["unqueezed_pad"]),
             make_node("Sub", [in_name, "unqueezed_pad"], ["Xsub"]),
             make_node("Pad", ["Xsub", "pad"], ["Xpadded"]),
             make_node("Add", ["Xpadded", "unqueezed_pad"], [out_name])]
    return nodes



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np

# Akida weights range is [-127, 127], int8
AKIDA_W_MAX = 127.0


def quantize_vector(x, scale, bitwidth=8, signed=True):
    # Decrease bitwidth when signed
    if signed:
        bitwidth -= 1
    # Project input
    proj_x = x * scale
    # Inputs needs to be quantized to bitwidth-bit
    epsilon = 2**-8
    x_int_bits = np.ceil(np.log2(np.abs(proj_x + epsilon)))
    qx_shift = np.maximum(0, x_int_bits - bitwidth).astype(np.int32)
    qx_8 = np.round(proj_x / (2 ** qx_shift)).astype(np.int32)
    # Rebuild quantized inputs
    return qx_8 << qx_shift


def quantize_weights(kernel):
    # Absolute max value calculated on first dim
    q_axis = tuple(range(1, kernel.ndim))
    abs_max = np.abs(kernel).max(q_axis)
    # Clip absolute max value
    abs_max = np.maximum(abs_max, 2.**-16)
    # Calculate float scale
    scale = AKIDA_W_MAX / abs_max
    # Project weights and bias to quantized int8 domain, but keeping them in float
    proj_kernel = kernel * align_to(scale, kernel.ndim, axis=0)
    # Now kernel can be finally be quantized to int8
    qkernel = np.round(proj_kernel)
    qkernel = np.clip(qkernel, -128, 127)
    qkernel = qkernel.astype(np.int8)
    return qkernel, scale


def fold_zero_point(bias, kernel, zero_point):
    # Fold zero point into bias.
    # To reduce quantization error, kernel should be quantized / dequantized previously.
    # Note this is possible if we project each zero point through
    # its respective axis in the weights (with FCXY format)
    zero_point = align_to(zero_point, kernel.ndim)
    q_axis = tuple(range(1, kernel.ndim))
    bias = bias - (zero_point.astype("float32") * kernel).sum(axis=q_axis)
    return bias


def align_to(x, ndims, axis=1):
    # Reshape x in order to have ndims, moving the information to a decided axis
    align_shape = [1 if axis != i else -1 for i in range(ndims)]
    return np.reshape(x, align_shape)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
quantizeml main command-line interface.
"""

import argparse
import json
import os
import re
import numpy as np
import os.path

from importlib import metadata

from keras.models import Model, Sequential

from . import load_model, save_model
from .layers import QuantizationParams
from .models import quantize, dump_config
from .models.transforms import insert_rescaling
from .analysis.cli import main as analysis_cli, add_analysis_arguments


def quantize_model(model_path, quant_config, qparams, samples, num_samples, batch_size, epochs,
                   output_name=None, quantize_until=None):
    """ CLI entry point to quantize a model using the provided configuration.

    Args:
        model_path (str): Path to the model to quantize.
            Only supported for Keras and ONNX models.
        quant_config (str): Path to the quantization configuration file.
        qparams (QuantizationParams): global quantization parameters.
        samples (str): calibration samples file path.
        num_samples (int): number of samples to use in the provided samples.
        batch_size (int): batch size for calibration.
        epochs (int): number of epochs for calibration.
        output_name (str, optional): name of the output quantized model. If none provided
            set a default name as <model>_<config>.h5. Defaults to None.
        quantize_until (str, optional): Layer/node name up to which to quantize. Defaults to None.
    """
    # Build name for the output model
    model_name = os.path.splitext(model_path)[0]
    _, ext = os.path.splitext(model_path)
    model_ext = ext.lower()

    # Set default output_name using the config when there is one and the qparams otherwise
    if output_name is None:
        if quant_config:
            output_name = f"{model_name}_{os.path.splitext(os.path.basename(quant_config))[0]}.h5"
        else:
            # Quantization suffix has a well-identified structure, only keep the actual base name
            # without it
            model_name = re.split(r'(_i\d_w\d_a\d)', model_name)[0]
            output_name = f"{model_name}_i{qparams.input_weight_bits}_w{qparams.weight_bits}"\
                          f"_a{qparams.activation_bits}{model_ext}"

    # Load the configuration file
    if quant_config:
        with open(quant_config) as f:
            config = json.load(f)
    else:
        config = None

    # Load the model
    model = load_model(model_path)

    # Load calibration samples
    if samples:
        data = np.load(samples)
        samples_arr = np.concatenate([data[item] for item in data.files])
        num_samples = len(samples_arr)
    else:
        samples_arr = None

    # Quantize the model and save it
    print(f"Quantizing model {model_path} with configuration file {quant_config}.")
    model_q = quantize(model, q_config=config, qparams=qparams, samples=samples_arr,
                       num_samples=num_samples, batch_size=batch_size, epochs=epochs,
                       quantize_until=quantize_until)
    output_name = save_model(model_q, output_name)
    print(f"Saved quantized model to {output_name}.")


def dump_model_config(model_path, output_name=None):
    """ CLI entry point to dump the quantization configuration from a model.

    Args:
        model_path (str): Path to the model to extract the configuration from.
            Only supported for Keras models.
        output_name (str): Path to save the configuration.
            Defaults to <model_path>_quant_config.json.
    """
    # Build name for the output model
    if output_name is None:
        model_name = os.path.splitext(model_path)[0]
        output_name = f"{model_name}_quant_config.json"

    # Load the model and get its quantization configuration
    model = load_model(model_path)
    # Raise an error if the model isn't a Keras
    if not isinstance(model, (Model, Sequential)):
        raise TypeError("Expecting a Keras model.")
    config = dump_config(model)
    with open(output_name, "w") as f:
        json.dump(config, f, indent=4)
    print(f"Saved quantization configuration to {output_name}.")


def insert_rescaling_and_save(model_path, dest_path, scale, offset):
    """ CLI entry point to insert a Rescaling layer in a model.

    Args:
        model_path (str): Path to the source model.
            Only supported for Keras models.
        dest_path (str): Path to the destination model.
        scale (float): the Rescaling scale
        offset (float): the Rescaling offset
    """
    model = load_model(model_path)
    # Raise an error if the model isn't a Keras
    if not isinstance(model, (Model, Sequential)):
        raise TypeError("Expecting a Keras model.")
    updated = insert_rescaling(model, scale, offset)
    updated.save(dest_path)


def main():
    """ CLI entry point.

    Contains an argument parser with specific arguments depending on the model to be created.
    Complete arguments lists available using the -h or --help argument.

    """
    # Define a strictly positive int type for parameters
    def positive_int(value):
        ivalue = int(value)
        if ivalue <= 0:
            raise argparse.ArgumentTypeError("%s is an invalid value, must be >0." % value)
        return ivalue

    parser = argparse.ArgumentParser()
    sp = parser.add_subparsers(dest="action")
    sp.add_parser("version", help="Display quantizeml version.")

    # Quantize arguments
    q_parser = sp.add_parser(
        "quantize", help="Quantize an input model, given a quantization configuration file.")
    q_parser.add_argument("-m", "--model", type=str, required=True, help="Model to quantize")
    q_parser.add_argument("-c", "--quantization_config", type=str, default=None,
                          help="Quantization configuration file")
    q_parser.add_argument("-a", "--activation_bits", type=positive_int, default=8,
                          help="Activation quantization bitwidth")
    q_parser.add_argument("--per_tensor_activations", action='store_true',
                          help="Quantize activations per-tensor")
    q_parser.add_argument("-w", "--weight_bits", type=positive_int, default=8,
                          help="Weight quantization bitwidth")
    q_parser.add_argument("-i", "--input_weight_bits", type=positive_int, default=8,
                          help="Input layer weight quantization bitwidth")
    q_parser.add_argument("-b", "--buffer_bits", type=positive_int, default=32,
                          help="Buffer quantization bitwidth")
    q_parser.add_argument("-s", "--save_name", type=str,
                          help="Name for saving the quantized model.")
    q_parser.add_argument("-sa", "--samples", type=str, default=None,
                          help="Set of samples to calibrate the model (.npz file).")
    q_parser.add_argument("-ns", "--num_samples", type=positive_int, default=1024,
                          help="Number of samples to use for calibration, only used when 'samples'"
                               " is not provided.")
    q_parser.add_argument("-bs", "--batch_size", type=positive_int, default=32,
                          help="Batch size for calibration.")
    q_parser.add_argument("-e", "--epochs", type=positive_int, default=1,
                          help="Number of epochs for calibration.")
    q_parser.add_argument("-qu", "--quantize_until", type=str, default=None,
                          help="Layer/node name up to which to quantize. "
                          "By default the whole model is quantized.")

    # Dump config arguments
    c_parser = sp.add_parser(
        "config", help="Extract quantization configuration from a Keras model.")
    c_parser.add_argument("-m", "--model", type=str, required=True,
                          help="Model to extract config from.")
    c_parser.add_argument("-o", "--output_path", type=str, help="Store quantization configuration. "
                          "Defaults to <model>_quant_config.json")

    # insert_rescaling action and arguments
    ir_parser = sp.add_parser(
        "insert_rescaling", help="Insert a Rescaling layer at the beginning of the Keras model.")
    ir_parser.add_argument("-m", "--model", type=str, required=True,
                           help="Path to the source Model")
    ir_parser.add_argument("-d", "--dest_model", type=str, required=True,
                           help="Path to the destination Model")
    ir_parser.add_argument("-s", "--scale", type=float, required=True,
                           help="The Rescaling scale")
    ir_parser.add_argument("-o", "--offset", type=float, required=True,
                           help="The Rescaling offset")

    # Analysis actions
    a_parser = sp.add_parser("analysis",
                             help="Tool set to analyze potential models to be quantized")
    add_analysis_arguments(a_parser)

    args = parser.parse_args()

    if args.action == "version":
        print(metadata.version('quantizeml'))
    elif args.action == "quantize":
        qparams = QuantizationParams(activation_bits=args.activation_bits,
                                     per_tensor_activations=args.per_tensor_activations,
                                     weight_bits=args.weight_bits,
                                     input_weight_bits=args.input_weight_bits,
                                     buffer_bits=args.buffer_bits)
        quantize_model(
            model_path=args.model,
            quant_config=args.quantization_config,
            qparams=qparams,
            samples=args.samples,
            num_samples=args.num_samples,
            batch_size=args.batch_size,
            epochs=args.epochs,
            output_name=args.save_name,
            quantize_until=args.quantize_until,
        )
    elif args.action == "config":
        dump_model_config(
            model_path=args.model,
            output_name=args.output_path,
        )
    elif args.action == "insert_rescaling":
        insert_rescaling_and_save(args.model, args.dest_model, args.scale, args.offset)
    elif args.action == "analysis":
        analysis_cli(args)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["QuantizationParams", "get_quantization_params", "quantization"]

from contextlib import contextmanager


class QuantizationParams:
    """ Class that holds quantization parameters.

    This is a read-only data class.

    Args:
        activation_bits (int, optional): activations quantization bitwidth. Defaults to 8.
        per_tensor_activations (bool, optional): whether to quantize activation per-tensor or
            per-axis. Defaults to False.
        weight_bits (int, optional): weights quantization bitwidth. Defaults to 8.
        output_bits (int, optional): outputs quantization bitwidth. Defaults to 8.
        input_weight_bits (int, optional): weights quantization bitwidth for the first layer.
            Defaults to 8.
        buffer_bits (int, optional): maximal buffer bitwidth allowed in operations.
            Defaults to 32.
    """

    def __init__(self, activation_bits=8, per_tensor_activations=False, weight_bits=8,
                 output_bits=8, input_weight_bits=8, buffer_bits=32):
        self._activation_bits = activation_bits
        self._per_tensor_activations = per_tensor_activations
        self._weight_bits = weight_bits
        self._output_bits = output_bits
        self._input_weight_bits = input_weight_bits
        self._buffer_bits = buffer_bits

    @property
    def activation_bits(self):
        return self._activation_bits

    @property
    def per_tensor_activations(self):
        return self._per_tensor_activations

    @property
    def weight_bits(self):
        return self._weight_bits

    @property
    def output_bits(self):
        return self._output_bits

    @property
    def input_weight_bits(self):
        return self._input_weight_bits

    @property
    def buffer_bits(self):
        return self._buffer_bits

    def __repr__(self) -> str:
        return (
            f"QuantizationParams("
            f"activation_bits={self.activation_bits}, "
            f"per_tensor_activations={self.per_tensor_activations}, "
            f"weight_bits={self.weight_bits}, "
            f"output_bits={self.output_bits}, "
            f"input_weight_bits={self.input_weight_bits}, "
            f"buffer_bits={self.buffer_bits})"
        )

    def __str__(self) -> str:
        return (
            f"Activation bits: {self.activation_bits}, "
            f"Per tensor activations: {self.per_tensor_activations}, "
            f"Weight bits: {self.weight_bits}, "
            f"Output bits: {self.output_bits}, "
            f"Input weight bits: {self.input_weight_bits}, "
            f"Buffer bits: {self.buffer_bits}"
        )


_quantization = QuantizationParams()


def get_quantization_params():
    """ Returns global quantization parameters.

    Returns:
        QuantizationParams: the quantization parameters
    """
    return _quantization


@contextmanager
def quantization(qparams):
    """ Sets quantization parameters in a context.

    Args:
        qparams (QuantizationParams): quantization parameters
    """
    # Use of global parameters
    global _quantization
    previous_state = _quantization
    try:
        # Set provided values
        _quantization = qparams
        yield
    finally:
        # Restore previous state
        _quantization = previous_state



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from collections import namedtuple
import warnings
import onnx
from .calibration import calibrate
from .register_patterns import PATTERNS_MAP, CUSTOM_PATTERNS_MAP
from .model import ONNXModel
from ..graph_tools import infer_partial_io, generate_node_names
from .. import layers as onnx_qlayers
from .transforms import sanitize
from ...layers.quantization_params import QuantizationParams
from ..layers.base_layer import _GLOBAL_REQUIRES_FP_INPUTS, _GLOBAL_REQUIRES_DOWNSCALE

# Define named tuples for QuantizerPattern and Quantizer
Quantizer = namedtuple('Quantizer', ['qlayer', 'qinputs', 'out_name'])


def search_block_from_node(node, model, quantize_until=None):
    """Try to search a quantizable node sequence from a target node.

    Args:
        node (NodeProto): the search start node.
        model (ONNXModel): the model containing the node.
        quantize_until (str, optional): if provided, limit the search until a node
            whose output matches to it. Defaults to None.

    Returns:
        list: a quantizable sequence of nodes. ``None`` if none found.
    """
    for qpattern in CUSTOM_PATTERNS_MAP + PATTERNS_MAP:
        pattern = qpattern.pattern
        # Try to recognize a sequence of nodes equal to the pattern size
        block_nodes = [node]
        for _ in range(len(pattern) - 1):
            # End search if quantize_until is provided
            if block_nodes[-1].output[0] == quantize_until:
                break
            outbound_nodes = model.get_children(block_nodes[-1])
            # A valid sequence cannot contain multiple outbounds.
            if len(outbound_nodes) != 1:
                break
            # Add outbound in the list of nodes
            block_nodes.extend(outbound_nodes)
        block_id = tuple(node.op_type for node in block_nodes)
        if block_id == pattern:
            # The node search ends because a pattern matchs with block_id.
            # Return all quantized functions that match with the same id.
            target_fn_list = []
            for qpattern in CUSTOM_PATTERNS_MAP + PATTERNS_MAP:
                if qpattern.pattern == block_id:
                    target_fn_list.extend(qpattern.f)
            return block_nodes, target_fn_list


def quantize_calibrated(model, tensors_range, quantize_until=None):
    """
    Given a calibrated onnx model and associated tensor ranges, create a quantized onnx
    model compatible with Brainchip's Akida IP and returns it as a new onnx model.

    Args:
        model: model to quantize
        tensors_range: dictionary of tensor name and its range.
            Range is a tuple of min and max values.
            Example: {"input_0": (-1.23, +4.56)}
        quantize_until (str, optional): quantization is performed until this node.
            Defaults to None.

    Returns:
        quantized onnx model.
    """
    assert isinstance(model, ONNXModel)

    # Reject multi-input-output models (yet)
    if len(model.input) != 1 or len(model.output) != 1:
        raise RuntimeError("Only single input/output models are supported.")
    if quantize_until and not any(node.name == quantize_until for node in model.nodes()):
        raise ValueError(f"'{quantize_until}' is not a recognized node in "
                         f"{model.graph().name}")

    # Rename operations to match with patterns
    graph = model.graph()

    # Copy target model to build a submodel with the remaining nodes (no quantizable)
    # and create an empty ONNXModel to build the new quantized model
    model_name = model.name or "quantized_model"
    qmodel = ONNXModel(onnx.helper.make_model(graph=onnx.GraphProto(name=model_name)))
    qmodel.input.extend(model.input)
    qmodel.set_opset_import(onnx_qlayers.DOMAIN, onnx_qlayers.VERSION)

    # Split nodes in blocks.
    # Note: list(nodes) clone each NodeProto in graph.node field
    remaining_nodes = list(model.nodes())

    # Start with all nodes linked to the input
    node_queue = []
    output_names = []
    qmodel.quantizers = []

    for node in model.nodes():
        if model.get_node_inputs(node) == [graph.input[0].name]:
            node_queue.append(node)
    while len(node_queue) > 0:
        # Process the first node in the queue
        target_node = node_queue.pop(0)
        # Creates a Quantizer if a sequence of nodes is found
        if block_qpattern := search_block_from_node(target_node, model, quantize_until):
            block_nodes, qpattern_func = block_qpattern
            # Each pattern is associated with a list of functions, prioritized by order.
            # The process tries the first function, and if it raises an exception,
            # it moves to the next one.
            error_trace = []
            for qlayer_func in qpattern_func:
                # Reset qlayer to avoid taking values from past iterations
                qlayer = None
                try:
                    # Initialize quantized layer
                    qlayer = qlayer_func(block_nodes, graph)
                    break
                except Exception as e:
                    # Try to quantize block nodes with next func
                    error_trace.append(f"{e.__class__.__name__}: {str(e)}")
                    continue

            if not isinstance(qlayer, onnx_qlayers.OnnxLayer):
                raise RuntimeError(f"Impossible to quantize {block_nodes}: it produces {qlayer} "
                                   f"which is not a valid {onnx_qlayers.OnnxLayer} object. \n"
                                   f"Error trace: \n\t* " + '\n\t* '.join(error_trace))

            # Find out if the inputs were quantized. If not, ignore the sequence
            try:
                qinputs = []
                for x in model.get_node_inputs(target_node):
                    if x == model.input[0].name:
                        # Skip nodes that are linked to inputs (e.g. InputQuantizer)
                        continue
                    quantizer_id = output_names.index(x)
                    qinputs.append(qmodel.quantizers[quantizer_id].qlayer)

            except ValueError:
                continue

            # Create intermediate Quantizer representation
            new_quantizer = Quantizer(qlayer, qinputs, block_nodes[-1].output[0])
            qmodel.quantizers.append(new_quantizer)
            # When quantize_until is provided, exclude output to stop quantization
            # for next nodes
            if new_quantizer.out_name != quantize_until:
                output_names.append(new_quantizer.out_name)
            else:
                output_names.append(None)
            # Remove nodes to be quantized from target model
            for node in block_nodes:
                remaining_nodes.remove(node)
            # Include in the queue the block children
            # Note get_children returns nodes in topological order, therefore a node
            # with multiple inputs will be processed once they have been quantized
            for child_node in model.get_children(block_nodes[-1]):
                if child_node not in node_queue:
                    node_queue.append(child_node)

    # No pattern was found if there is only InputQuantizer in quantizers list
    if len(qmodel.quantizers) <= 1:
        raise RuntimeError("No quantizable pattern found")

    # Compute each remaining input (those whose nodes are disconnected remaining model)
    partial_float_in, _ = infer_partial_io(remaining_nodes,
                                           exclude=list(model.get_initializer_name_set()))

    # Output needs to be dequantized when there are no remaining_nodes
    if len(remaining_nodes) == 0:
        partial_float_in.append(graph.output[0].name)

    # Main loop: quantize qlayers and concatenate them in qnodes
    value_info = qmodel.graph().value_info
    for qidx, quantizer in enumerate(qmodel.quantizers):
        last_quantizer = quantizer.out_name in partial_float_in
        # Note downscale is just implemented for QuantizedAdd and QuantizedDense1D
        downscale = not last_quantizer or \
            isinstance(quantizer.qlayer, tuple(_GLOBAL_REQUIRES_DOWNSCALE))
        # QuantizedAdd constraint: power-of-two input scales are mandatory.
        # In other words, we force output scale to be a power-of-two if target node
        # has at least one QuantizedAdd outbound.
        force_fp = False
        for q in qmodel.quantizers[qidx + 1:]:
            if quantizer.qlayer in q.qinputs:
                force_fp = isinstance(q.qlayer, tuple(_GLOBAL_REQUIRES_FP_INPUTS))
        # Build node previous to quantization in order to assing a custom output name
        # Note that this is not the case for InputQuantizer, since it is a new node in the graph,
        # it must have a new name.
        if not isinstance(quantizer.qlayer, onnx_qlayers.InputQuantizer):
            input_vi = [qi.output for qi in quantizer.qinputs]
            quantizer.qlayer.build(*input_vi, out_name=quantizer.out_name, downscale=downscale)
        # Quantize node to retrieve NodeProto and list of TensorProto (weights)
        out_tensor_ranges = tensors_range[quantizer.out_name]
        qnode, onnx_weights = quantizer.qlayer.quantize(*quantizer.qinputs,
                                                        out_tensor_range=out_tensor_ranges,
                                                        force_fp=force_fp)
        # Include new quantized node into qmodel
        qmodel.add_node(qnode)
        value_info.append(quantizer.qlayer.output)
        if len(onnx_weights) > 0:
            qmodel.initializer_extend(onnx_weights)
        # Update output name list
        output_names[qidx] = qnode.output[0]

    # Plug a dequantizer per each remaining input
    io_deq_map = []
    for iname in partial_float_in:
        # Create the respective dequantizer for each remaining input
        qlayer = qmodel.quantizers[output_names.index(iname)].qlayer
        deq = onnx_qlayers.Dequantizer(name=f"{qlayer.output.name}/dequantize")
        qnode, onnx_weights = deq.quantize(qlayer)
        qmodel.add_node(qnode)
        qmodel.initializer_extend(onnx_weights)
        # Make Dequantizer an output of the quantized model and an input of the remaining model
        qmodel.output.append(deq.output)
        # Save input output map (to merge submodels)
        io_deq_map.append((deq.output.name, iname))

    # Register functions in the quantized graph
    qmodel.model.functions.extend(onnx_qlayers.AKIDA_ONNX_LAYERS)

    # Finally build the quantized model
    if len(remaining_nodes) > 0:
        if quantize_until is None:
            warnings.warn("The following nodes were not quantized because their pattern "
                          "was not found in the scope: "
                          f"{[f'{x.name} ({x.op_type})' for x in remaining_nodes]}.")
        # Extract remaining submodel
        extractor = onnx.utils.Extractor(model.model)
        remaining_model = extractor.extract_model(partial_float_in, [model.output[0].name])
        # merge_models throws an error if it finds value_info name overlap in both models.
        # To avoid it, we give priority to those present in the quantized model.
        for value_info in list(remaining_model.graph.value_info):
            if qmodel.find_value_info_by_name(value_info.name):
                remaining_model.graph.value_info.remove(value_info)
        # Use onnx.compose helper tool to merge the models manually,
        # avoiding some issues (e.g. topological ordering).
        qmodel = onnx.compose.merge_models(qmodel.model, remaining_model, io_map=io_deq_map)
        qmodel = ONNXModel(qmodel)
    return qmodel


def quantize(model,
             qparams=QuantizationParams(),
             samples=None,
             num_samples=1024,
             batch_size=None,
             quantize_until=None):
    """
    Given an onnx model and calibration data reader, create a quantized onnx
    model compatible with Brainchip's Akida IP and returns it as a new onnx model.

    Args:

        model (ModelProto): the onnx model instance to quantize
        qparams (QuantizationParams, optional): Quantization parameters. It is used
            to determine if quantizing per-tensor or per-axis.
        samples (list of numpy arrays, optional): List of input samples to use for
            calibration. If not provided, random samples will be generated. Defaults
            to None.
        num_samples (int, optional): Number of samples to use for calibration.
            Defaults to 1024.
        batch_size (int, optional): Batch size to use for calibration. Defaults to
            None.
        quantize_until (str, optional): name of the node until which to quantize:
            other nodes after it will stay unchanged. Defaults to None.

    Returns:
        quantized onnx model.
    """
    # For now only a limited QuantizationParams configuration is supported: test that
    if (
            qparams.activation_bits != 8 or
            qparams.buffer_bits != 32 or
            qparams.input_weight_bits != 8 or
            qparams.output_bits != 8 or
            qparams.weight_bits != 8):
        raise ValueError("Only default bitwidth params params qparams is allowed.")

    # Generate node names if needed
    generate_node_names(model)

    # Parse ModelProto into a ONNXModel
    onnx_model = ONNXModel(model)

    # Sanitize the input model
    onnx_model = sanitize(onnx_model)

    # Compute statistical ranges
    tensors_range = calibrate(onnx_model.model,
                              samples,
                              num_samples,
                              batch_size,
                              per_tensor_activations=qparams.per_tensor_activations)

    # Quantize model with tensors_range
    qmodel = quantize_calibrated(onnx_model, tensors_range, quantize_until=quantize_until)
    return qmodel.model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["calibrate", "calibration_required"]

import numpy as np
import tensorflow as tf

from copy import deepcopy
from keras.layers import (Conv2D, Conv2DTranspose, SeparableConv2D, DepthwiseConv2D, Dense, Dropout,
                          Reshape, Flatten, MaxPool2D, GlobalAvgPool2D)

from .transforms.insert_layer import insert_in_config
from .transforms.transforms_utils import get_layers_by_type
from .utils import apply_weights_to_model, get_model_input_dtype
from ..layers import (OutputQuantizer, OutputObserver, PaddedConv2D, DepthwiseConv2DTranspose,
                      BufferTempConv, DepthwiseBufferTempConv, StatefulRecurrent,
                      StatefulProjection, Dequantizer)
from ..tensors import FixedPoint, floor_log2


def calibration_required(model):
    """Checks if a model requires calibration.

    If one of the 'OutputQuantizer' layers in the model has its range_max variable set to 1, it
    requires calibration.

    Args:
        model (keras.Model): the model to check

    Returns:
        bool: True if calibration is required, False otherwise.
    """
    calibrables = get_layers_by_type(model, OutputQuantizer)
    for calibrable in calibrables:
        # If the model has never been calibrated, all range_max of the OutputQuantizer objects
        # will be set to 1.
        if tf.reduce_all(calibrable.variables[0] == 1):
            return True
    # all calibrable objects are set
    return False


def _get_calibration_model(model, qmodel):
    """Builds a calibration model with OutputObserver added between blocks.

    Args:
        model (keras.Model): input model
        qmodel (keras.Model): quantized model

    Returns:
        keras.Model, dict: the calibration ready model and dict mapping end of block layer names to
        the name of their OutputObserver.
    """
    # Get model config to edit
    config = deepcopy(model.get_config())

    # Insert OutputObservers where OutputQuantizers are
    end_of_blocks = {}
    for layer in qmodel.layers:
        if getattr(layer, 'out_quantizer', None):
            # Build an observer and store it for future use
            observer = OutputObserver(layer.out_quantizer.axis)
            end_of_blocks[layer.name] = observer.name
            # Insert observer in config
            insert_in_config(model, layer.name, observer, config)

    # Build the calibration model from the config
    calibration_model = model.from_config(config)

    # Load original weights
    variables_dict = {var.name: var for var in model.variables}
    apply_weights_to_model(calibration_model, variables_dict, False)
    return calibration_model, end_of_blocks


def _get_next_layer(layer, supported_layers, skippable_layers=(), reshape_ops=[]):
    """ Finds the layer following a target layer.

    Args:
        layer (keras.Layer): the layer of interest
        supported_layers (tuple): layer types that will support equalization.
        skippable_layers (tuple, optional): layer types that can be skipped. Defaults to ().
        reshape_ops (list, optional): list of reshape operation encountered as
            [(input_shape,), (output_shape,)]. Defaults to [].

    Returns:
        keras.Layer: the layer following the layer of interest if valid, None otherwise
    """
    # Limit support to single outbound
    outbounds = layer.outbound_nodes
    if len(outbounds) != 1:
        return None
    next_layer = outbounds[0].layer

    # StatefulProjection with downshape or upshape is not supported by CLE
    if isinstance(next_layer, StatefulProjection) and (next_layer.downshape or next_layer.upshape):
        return None
    # If the layer is supported, it is a valid candidate
    if isinstance(next_layer, supported_layers):
        return next_layer
    # If the next layer can be skipped, recursively call the function
    elif isinstance(next_layer, skippable_layers):
        return _get_next_layer(next_layer, supported_layers, skippable_layers, reshape_ops)
    # If the next layer is a Reshape or Flatten store the performed ops and continue (limit reshape
    # ops to a single layer)
    elif isinstance(next_layer, (Reshape, Flatten)) and len(reshape_ops) == 0:
        reshape_ops.append([next_layer.input_shape[1:], next_layer.output_shape[1:]])
        return _get_next_layer(next_layer, supported_layers, skippable_layers, reshape_ops)

    # If next layer is not supported, alignment cannot happen
    return None


def _set_and_equalize(qmodel, layer_name, range_max):
    """ Set 'layer_name' output quantizer range_max to the ideal value.

    The ideal range_max is computed from the maximum value the target layer OutputQuantizer's
    can represent given it's bitwidth, and the float calibrated range_max.

    The ratio between the calibrated value and the ideal value is stored in the OutputQuantizer and
    will be applied when the scale_out operation happens. Equalization happens by dividing weights
    of the next layer by the ratio to preserve global outputs.

    Args:
        qmodel (keras.Model): quantized keras model
        layer_name (str): layer name where to set the range_max
        range_max (tf.Tensor): the float calibrated range_max
    """
    # First set the calibrated range_max in the target OutputQuantizer
    target_layer = qmodel.get_layer(layer_name)
    target_layer.out_quantizer.range_max.assign(range_max)

    # Skip layers that only perform out_shift since rescaling rate needs a scale out to be applied
    if not hasattr(target_layer.out_quantizer, 'qscales'):
        return

    # Define layers that will support or can be skipped when equalizing
    supported_layers = (Conv2D, PaddedConv2D, Conv2DTranspose, SeparableConv2D, DepthwiseConv2D,
                        Dense, DepthwiseConv2DTranspose, BufferTempConv, DepthwiseBufferTempConv)
    skippable_layers = (Dropout, MaxPool2D, GlobalAvgPool2D, Dequantizer)
    # When quantized per-tensor, Reshaping layers can be skipped as the rescaling rate single value
    # will be used during equalization on all weights values
    if target_layer.out_quantizer.axis == 'per-tensor':
        skippable_layers += (Reshape, Flatten)

    # Retrieve next_layer: if there is no candidate layer following layer_name, equalization cannot
    # happen
    reshape_ops = []
    next_layer = _get_next_layer(target_layer, supported_layers, skippable_layers, reshape_ops)
    # For now cross layer equalization for (Dephwise)BufferTempConvLayers is possible only if their
    # input is quantized per-tensor.
    if next_layer is None or (isinstance(next_layer, (BufferTempConv, DepthwiseBufferTempConv))
                              and target_layer.out_quantizer.axis == 'per-axis'):
        return

    # Compute ideal range_max
    bitwidth = target_layer.out_quantizer.bitwidth
    frac_bits = bitwidth - floor_log2(range_max)
    int_max = FixedPoint.int_max(bitwidth)
    ideal_range_max = FixedPoint(int_max, bitwidth, frac_bits).to_float().numpy()

    # Check that values are different
    if np.all(ideal_range_max == range_max):
        return

    # Set ideal_range_max into the target OutputQuantizer
    target_layer.out_quantizer.range_max.assign(ideal_range_max)

    # range_max can be approximately 0 on some channels (eg. a ReLU where an input channel
    # had all negative values). When that happens, range_max set on output quantizer
    # (ideal_range_max) is exactly zero. In that case rescaling_rate is forced to 1
    # so that cross-layer equalization does nothing.
    if target_layer.out_quantizer.axis == 'per-axis':
        range_max = range_max.numpy()
        range_max[ideal_range_max == 0] = 1
        ideal_range_max[ideal_range_max == 0] = 1

    # Compute the rescaling rate
    rescaling_rate = ideal_range_max / range_max

    # Set rate in the target OutputQuantizer
    target_layer.out_quantizer.rescaling_rate.assign(rescaling_rate)

    # Divide weights of the next layer by the rescaling rate
    next_weights = next_layer.get_weights()
    new_weights = next_weights[0]

    # Apply reshape ops
    if reshape_ops:
        # Supporting a single reshape ops
        reshape_ops = reshape_ops[0]
        # Check that last dimension is unchanged by the reshape ops, if that's not the case, the op
        # needs to be undone on weights, eg. for Flatten: X*Y*C, F -> X, Y, C, F
        if reshape_ops[0][-1] != reshape_ops[1][-1]:
            F = (new_weights.shape[-1], )
            new_weights = np.reshape(new_weights, reshape_ops[0] + F)
        # Invert last two dimensions (input and output channels), eg. for flatten:
        # X, Y, C, F -> X, Y, F, C
        axes = list(range(0, len(new_weights.shape)))
        assert len(axes) > 1
        axes[-2], axes[-1] = axes[-1], axes[-2]
        new_weights = np.transpose(new_weights, axes)
    elif not isinstance(next_layer, Conv2DTranspose):
        # Expand dims to allow broadcasting on the expected dimension. Skipping Conv2dTranspose
        # since their kernel is (H, W, F, C), where C is the dimension of rescaling_rate
        rescaling_rate = np.expand_dims(rescaling_rate, -1)

    # Apply rescaling rate
    new_weights /= rescaling_rate

    # Undo reshape ops
    if reshape_ops:
        # Revert last dimensions swap, eg. for Flatten: X, Y, F, C -> X, Y, C, F
        new_weights = np.transpose(new_weights, axes)
        # Undo the reshape op on weights when last dimension is changed, eg. for Flatten:
        # X, Y, C, F -> X*Y*C, F
        if reshape_ops[0][-1] != reshape_ops[1][-1]:
            F = (new_weights.shape[-1], )
            new_weights = np.reshape(new_weights, reshape_ops[1] + F)

    next_weights[0] = new_weights
    next_layer.set_weights(next_weights)


def _warmup_calib(calibration_model, warmup_samples, end_of_blocks, batch_size):
    """ Helper method to calibrate the buffer TENN models with warmpup period.

    Args:
        calibration_model (keras.Model): the calibration model.
        warmup_samples (tf.Dataset, np.array or generator, optional): warmup samples.
        end_of_blocks (dict): dict mapping end of block layer names to the name of their
            OutputObserver.
        batch_size (int): calibration batch size.
    """

    # Fill the fifos.
    calibration_model.predict(x=warmup_samples, batch_size=batch_size)
    # reset the range_max
    for _, observer in end_of_blocks.items():
        range_max = calibration_model.get_layer(observer).range_max
        range_max.assign(tf.ones_like(range_max))


def _split_samples(samples, buffer_layers, batch_size):
    """Helper to split the samples into warmup and calibration samples for the TENN models

    Args:
        samples (np.array): samples.
        buffer_layers (list: keras.Layers): the model (Depthwise)BufferTempConv layers
        batch_size (int): calibration batch_size.

    Returns:
        np.array, np.array, int: warmup samples, effective calibration samples and warmup
        steps.
    """
    # Extract warmup_factor. The number of samples required to fill all model fifos.
    warmup_factor = 0
    for buf_layer in buffer_layers:
        warmup_factor += buf_layer.kernel_size
    if warmup_factor * batch_size > samples.shape[0]:
        raise ValueError(f"Number of samples must be greater than {warmup_factor} * batch_size.")
    # Extract warmup samples
    warmup_samples = samples[:warmup_factor * batch_size]
    # Extract calibration samples
    calib_samples = samples[warmup_factor * batch_size:]
    return warmup_samples, calib_samples, warmup_factor


def calibrate(model, qmodel, samples=None, num_samples=1024, batch_size=None, epochs=1):
    """Calibrates the model using the provided samples.

    With TENN models only np.array samples are supported for calibration. Those should have
    a temporally coherent data, which means that their expected shape is
    [batch_size*Seq, dim_0,, ..., dim_n] for spatiotemporal TENNs where:

        - batch_size is the same batch_size provided to the calibration.
        - Seq is a dataset parameter that defines the temporally coherent data (eg number of frames
          per video clips).

    and [batch_size, (model.input_shape)] for recurrent TENNs.

    When no samples are provided, random samples are generated.

    Args:
        model (keras.Model): the original model
        qmodel (keras.Model): the quantized model to calibrate
        samples (tf.Dataset, np.array or generator, optional): calibration samples. When no samples
            are provided, random samples are generated. Defaults to None.
        num_samples (int, optional): number of samples to use in the provided samples or number of
            samples to generate. Defaults to 1024.
        batch_size (int, optional): the batch size. Defaults to None.
        epochs (int, optional): the number of epochs. Defaults to 1.

    """
    if not calibration_required(qmodel):
        return

    # Disable sumbsampling in StatefulRecurrent layers before building the calibration model
    rec_layers = get_layers_by_type(model, StatefulRecurrent)
    for ly in rec_layers:
        ly.subsample_ratio = None

    # Build a calibration model which is a float model with OutputObservers at locations where the
    # quantized model has OutputQuantizers.
    calibration_model, end_of_blocks = _get_calibration_model(model, qmodel)

    if samples is None:
        # Generate random samples
        samples_shape = (num_samples, ) + qmodel.input_shape[1:]
        samples_type = get_model_input_dtype(model).as_numpy_dtype

        rng = np.random.default_rng()
        iinfo = np.iinfo(samples_type)
        samples = rng.integers(iinfo.min, iinfo.max, samples_shape, dtype=samples_type)

    # Extract number of (Depthwise)BufferTempConv layers
    buf_layers = get_layers_by_type(calibration_model, (BufferTempConv, DepthwiseBufferTempConv))
    if buf_layers or rec_layers:
        if not isinstance(samples, np.ndarray):
            raise TypeError("TENN models calibration is only possible with np.array samples."
                            f" Received {type(samples)}. Please convert them to a compatible"
                            " format before the calibration. \n"
                            " Spatiotemporal expects the following shape:"
                            " [batch_size*Seq, dim_0,, ..., dim_n] where:"
                            " - batch_size is the same batch_size provided to the calibration."
                            " - Seq is a dataset parameter that defines the temporally coherent"
                            " data (eg number of frames per video clips). \n"
                            " Recurrent expects [batch_size, (model.input_shape)] array.")
        if buf_layers:
            # With ST TENN models, the number of samples is infered from the provided samples
            num_samples = len(samples)
            if batch_size is None:
                raise ValueError("batch_size not specified. With spatiotemporal TENN models "
                                 "batch_size must be given for the calibration, otherwise the "
                                 "result is unpredictible. Refer to 'calibrate' function docstring "
                                 "for more details.")

            # The number of samples must be a multiple of the batch_size to prevent a reset of the
            # FIFO
            if num_samples % batch_size != 0:
                raise ValueError("When calibrating spatiotemporal TENN, num_samples must be a "
                                 f"multiple of batch_size, got num_samples={num_samples} and "
                                 f"batch_size={batch_size}.")

    # Compute step value otherwise 'predict' will run until samples are exhausted (ie. indefinitely
    # if samples is a dataset with repeat enabled)
    if batch_size is None:
        steps = num_samples
    else:
        assert batch_size > 0, "The batch size should be strictly positive."
        steps = np.ceil(num_samples / batch_size)

    for i in range(epochs):
        if rec_layers:
            # Ensure graph execution
            model_func = tf.function(calibration_model)
            in_shape = calibration_model.input_shape[1]
            # Allow to "cut" the calibration samples in portions compatible with
            # the stateful model input shape.
            assert samples.shape[1] % in_shape == 0, (
                f"Each sample length {samples.shape[1]} "
                f"should be a multiple of the model input size {in_shape}.")
            in_step = samples.shape[1] // in_shape
            for i in range(in_step):
                model_func(samples[:, i * in_shape:(i + 1) * in_shape])
        else:
            # Set default calib samples and number of steps
            calib_samples = samples
            calibration_steps = steps
            # If the model is a buffer TENN, it normally requires a warmup period on the first epoch
            # in order to calibrate it. For that purpose batch_size is mandatory.
            if len(buf_layers) > 0 and batch_size and i == 0:
                warmup_samples, calib_samples, warmup_steps = _split_samples(samples, buf_layers,
                                                                             batch_size)
                _warmup_calib(calibration_model, warmup_samples, end_of_blocks, batch_size)
                calibration_steps = steps - warmup_steps
            calibration_model.predict(
                x=calib_samples, steps=calibration_steps, batch_size=batch_size)

    # Update quantized model OutputQuantizers range_max using OutputObservers calibrated values
    for eob, observer in end_of_blocks.items():
        _set_and_equalize(qmodel, eob, calibration_model.get_layer(observer).range_max)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["SMAPE", "Saturation", "print_metric_table"]

from collections import defaultdict
import warnings
import numpy as np
import keras


class SMAPE(keras.metrics.Metric):
    """Compute the Symmetric Mean Absolute Percentage Error (SMAPE) as:

    >>> mean(abs(x - y) / (abs(x) + abs(y)))

    Reference: https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error

    Args:
        name (str, optional): name of the metric. Defaults to "smape".
    """

    def __init__(self, name="smape", **kwargs):
        super().__init__(name=name, **kwargs)
        self.error = self.add_weight(name='error', initializer='zeros', dtype="float64")
        self.count = self.add_weight(name='count', initializer='zeros', dtype="float64")

    def update_state(self, y_true, y_pred):
        assert y_true.shape == y_pred.shape
        total = y_true.size

        # Skip values that undefine the metric
        mask = (y_true == 0) & (y_pred == 0)
        y_true = y_true[~mask]
        y_pred = y_pred[~mask]

        # Compute smape
        smape = np.sum(np.abs(y_true - y_pred) / (np.abs(y_true) + np.abs(y_pred)))
        self.error.assign_add(smape)

        # Update the metric count.
        # Note here we take into account the set of values y_true = y_pred = 0,
        # since the error they contribute is zero.
        self.count.assign_add(total)

    def result(self):
        if self.count == 0:
            return 0.0
        return self.error / self.count

    def reset_states(self):
        self.error.assign(0.0)
        self.count.assign(0.0)


class Saturation(keras.metrics.Metric):
    """Returns the percentage of saturating values.

    We consider a value saturated if it is one of {min_value, max_value}

    Args:
        min_value (np.ndarray, optional): the minimum of values.
            If not provided, it is inferred from the values type. Defaults to None.
        max_value (np.ndarray, optional): the maximum of values.
            If not provided, it is inferred from the values type. Defaults to None.
    """

    def __init__(self, name="saturation", min_value=None, max_value=None, **kwargs):
        super().__init__(name=name, **kwargs)
        self._min_value = min_value
        self._max_value = max_value
        self.total = self.add_weight(name='total', initializer='zeros', dtype="int64")
        self.count = self.add_weight(name='count', initializer='zeros', dtype="int64")

    @property
    def min_value(self):
        if self._min_value is None:
            iinfo = np.iinfo(self.dtype) if np.issubdtype(self.dtype, np.integer) else None
            assert iinfo, f"Unknown minimum value for data type {self.dtype}"
            return iinfo.min
        return self._min_value

    @property
    def max_value(self):
        if self._max_value is None:
            iinfo = np.iinfo(self.dtype) if np.issubdtype(self.dtype, np.integer) else None
            assert iinfo, f"Unknown maximum value for data type {self.dtype}"
            return iinfo.max
        return self._max_value

    def update_state(self, values):
        if np.any(values > self.max_value) or np.any(values < self.min_value):
            warnings.warn(f"Saturation is not accurate: there are values outside of range "
                          f"[{self.min_value}, {self.max_value}].")
        sat_values = np.sum((values == self.min_value) | (values == self.max_value))
        self.total.assign_add(sat_values)
        self.count.assign_add(values.size)

    def result(self):
        count = self.count if self.count != 0 else 1
        return 100 * (self.total / count)

    def reset_states(self):
        self.total.assign(0)
        self.count.assign(0)

    def get_config(self):
        config = super().get_config()
        config.update({"min_value": self._min_value, "max_value": self._max_value})
        return config


def print_metric_table(summary, model_name=""):
    """Print a table with the results of a set of metrics.

    The following format is expected:

    .. code-block::

        # Format for metrics
        # 1. Simple set of metrics
        metrics_for_key_1 = {"metric_1": key1_metric1_value, "metric_2": key1_metric2_value}
        # 2. List of simple set of metrics
        metrics_for_key_2 = [{"metric_1": key2_metric1_value1, "metric_2": key2_metric2_value1},
                             {"metric_1": key2_metric1_value2, "metric_2": key2_metric2_value2}]
        # 3. List of complex set of metrics
        metrics_for_key_3 = [metrics_for_key_2, metrics_for_key_2]

        # Summary
        summary = {
            "key_1": metrics_for_key_1,
            "key_2": metrics_for_key_2,
            "key_3": metrics_for_key_3,
        }

    Args:
        summary (dict): summary of metrics to draw
        model_name (str, optional): A model name to display. Defaults to "".

    Note:
        All metrics must contain the same set of measures
    """
    FLOAT_PRECISION = 4
    saturate = False

    def _flat_measures(summary):
        nonlocal saturate
        new_summary = {}
        for name, measures in summary.items():
            if isinstance(measures, (list, tuple)):
                new_el = _flat_measures({f"{name}:{idx+1}": v for idx, v in enumerate(measures)})
            else:
                assert isinstance(measures, dict), f"Wrong measurement format for {name}."
                # Check if saturation is higher than 60 %
                saturated_layer = measures.get("Saturation (%)", 0.0) > 60
                new_el = {("*" if saturated_layer else "") + name: measures}
                saturate |= saturated_layer
            new_summary.update(new_el)
        return new_summary

    def _find_max_length(summary):
        max_key_len = defaultdict(int)
        for values in summary.values():
            for key, val in values.items():
                # Variable length is given by: integer part + '.' + float precision
                var_len = len(str(val).split('.')[0]) + 1 + FLOAT_PRECISION
                # Take into account length of key
                max_key_len[key] = max(max_key_len[key], var_len, len(key))
        return max_key_len

    summary = _flat_measures(summary)
    # Compute lengths per column
    title_names = "Layer/node"
    max_name_len = max(*(len(x) for x in summary), len(title_names))
    max_len_per_key = _find_max_length(summary)

    # Print head
    title_measures_cols = " | ".join([k.ljust(v) for k, v in max_len_per_key.items()])
    title_cols = title_names.ljust(max_name_len) + " | " + title_measures_cols
    total_table_len = len(title_cols)
    print(f"\nQuantization error for {model_name}:\n" + "=" * total_table_len)
    print(title_cols + "\n" + "=" * total_table_len, end="")

    # Print measures
    for name, measures in summary.items():
        print("\n" + name.ljust(max_name_len), end="")
        for key, just_len in max_len_per_key.items():
            if key not in measures:
                raise KeyError(f"'{name}' does not contain '{key}' in measures.")
            str_measure = ("{:." + str(FLOAT_PRECISION) + "f}").format(measures[key])
            print(" | " + str_measure.ljust(just_len), end="")
    print("\n" + "=" * total_table_len)

    # Print footer
    if saturate:
        print("* Measurement results are not reliable, as outputs saturate by more than 60%")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["untranspose_gemm_weights"]

import onnx
import onnx.numpy_helper

from ...graph_tools import get_field, replace_field
from ..model import ONNXModel


def untranspose_gemm_weights(model):
    """Retransposes gemm weights if they are transposed.

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    assert isinstance(model, ONNXModel)

    for node in model.nodes():
        if node.op_type == "Gemm":
            transB = get_field(node, "transB", -1)
            if transB in (-1, 0):
                # Transpose weights
                weights = model.get_variable(node.input[1])
                weights = weights.T

                # Update Tensor proto
                weights_tp = model.get_initializer(node.input[1])
                weights_tp.CopyFrom(onnx.numpy_helper.from_array(weights, node.input[1]))

                # Update or create attribute value
                if transB == -1:
                    node.attribute.append(onnx.helper.make_attribute(key="transB", value=1))
                else:
                    replace_field(node, "transB", 1)
    model.check_model()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np
import onnxruntime

from ...onnx_support.quantization.quantize import ONNXModel
from ...onnx_support.graph_tools import get_field
from ...onnx_support.quantization.weights import align_to
from ..tools import generate_onnx_random_samples, extract_submodel_from_qnode
from .common import compare, remove_outliers, eval_metrics, compute_saturation

skippable_qnodes = ("InputQuantizer", "Dequantizer")


def _is_measurable(qnode):
    return qnode.op_type not in skippable_qnodes and qnode.domain == "com.brainchip"


def _get_scale_zp(node, qmodel):
    scale, zp = get_field(node, "scale"), 0
    # Zero point could be different to zero if node is an InputQuantizer
    if node.op_type == "InputQuantizer":
        zp = qmodel.get_variable(node.input[2])
    return scale, zp


def _dequantize(x, scales, zp=0, otype="float32"):
    return ((x.astype("int32") - align_to(zp, x.ndim)) / align_to(scales, x.ndim)).astype(otype)


def _get_node(node_name, model):
    node = model.find_node_by_name(node_name)
    if node is None:
        raise RuntimeError(f"{model.name} must have node '{node_name}'")
    return node


def _compute_mask(x):
    min_value = np.iinfo(x.dtype).min
    max_value = np.iinfo(x.dtype).max
    mask = (x > min_value) & (x < max_value)
    return mask


def _search_quantized_target_nodes(model, target_node_name=None):
    if target_node_name is not None:
        target_nodes = [_get_node(target_node_name, model)]
    else:
        target_nodes = model.nodes()
    # Filter target nodes
    target_nodes = [node for node in target_nodes if _is_measurable(node)]
    if len(target_nodes) == 0:
        raise ValueError(f"{model.name} does not contain nodes that generate quantization error!")
    return target_nodes


def compare_outputs(foutputs, qoutputs, qscales, per_channel=False):
    """Measures the error in a set of arrays

    Args:
        foutputs (np.ndarray): the output of a float layer.
        qoutputs (np.ndarray): the quantized output to be compare with ``foutputs``.
        qscales (np.ndarray): scales to dequantize ``qoutputs``.
        per_channel (bool, optional): comparison is done for each channel. Defaults to False.

    Returns:
        dict or list: the quantization error.
    """
    assert np.issubdtype(qoutputs.dtype, np.integer), f"{qoutputs} must be integer"

    axis = (1 if foutputs.ndim > 1 else 0) if per_channel else None
    # Saturation is reduced through axis = 1 (format BCHW)
    saturation = compute_saturation(qoutputs, axis=axis)

    # Creates a mask where is indicated if a value saturate or not
    mask = _compute_mask(qoutputs)

    # Compare float/quantized outputs.
    # Note we need to dequantize qoutputs to be compared with foutputs
    # as well as excluding saturated values, since the error is ambiguous out of range.
    qoutputs = _dequantize(qoutputs, qscales)
    foutputs = remove_outliers(foutputs, mask, axis=axis)
    qoutputs = remove_outliers(qoutputs, mask, axis=axis)
    return compare(foutputs, qoutputs, saturation)


def quantization_error(fmodel, qmodel, target_node=None, batch_size=16, seed=None):
    """Measures the node quantization error in a set of ONNX models

    Args:
        fmodel (onnx.ModelProto): the float model.
        qmodel (onnx.ModelProto): the quantized version of `fmodel`.
        target_node (str, optional): computation error is performed only in the target node,
            expanding the analysis to each output channel. Defaults to None.
        batch_size (int, optional): the batch size. Defaults to 16.
        seed (int, optional): a random seed. Defaults to None.

    Returns:
        dict: the quantization error of the target nodes
    """
    fmodel = ONNXModel(fmodel)
    qmodel = ONNXModel(qmodel).clone()

    # Search target nodes and update graph outputs
    qtarget_nodes = _search_quantized_target_nodes(qmodel, target_node_name=target_node)

    # Create a quantized intermediate model with the inputs of qtarget_nodes
    qmodel.graph().ClearField("output")
    for qnode in qtarget_nodes:
        for iname in qnode.input:
            qvi = qmodel.find_value_info_by_name(iname)
            if qvi is not None and qvi not in qmodel.output:
                qmodel.output.append(qvi)

    # Retrieve all scales and zero points. Need to compute float inputs
    in_scales, in_zps, out_names = {}, {}, []
    outputs_to_node = qmodel.output_name_to_node()
    for qvi in qmodel.output:
        in_scales[qvi.name], in_zps[qvi.name] = _get_scale_zp(outputs_to_node[qvi.name], qmodel)
        out_names.append(qvi.name)

    # Generate a random set of samples
    samples = generate_onnx_random_samples(qmodel, batch_size=batch_size, seed=seed)

    # Compute quantization error per layer:
    # Generate the set of input quantized samples
    qinputs = onnxruntime.InferenceSession(qmodel.serialized).run(None, samples)
    qinputs = dict(zip(out_names, qinputs))

    # Compute quantization error per quantized node
    summary = {}
    per_channel = target_node is not None
    for qnode in qtarget_nodes:
        # Extract submodel from fmodel that represent the qnode operations in float-domain
        subfmodel = extract_submodel_from_qnode(qnode, qmodel, fmodel)

        # Extract a submodel from qmodel which contains just qnode
        subqmodel = extract_submodel_from_qnode(qnode, qmodel, qmodel)

        # Build feeds
        # Note: input names may change given qmodel has new nodes (e.g. InputQuantizer)
        qfeed = {k: qinputs[k] for k in [vi.name for vi in subqmodel.input]}
        ffeed = {k: _dequantize(qinputs[qk], in_scales[qk], zp=in_zps[qk])
                 for qk, k in zip(qfeed.keys(), [x.name for x in subfmodel.input])}

        # Forward inputs in submodels
        foutputs = onnxruntime.InferenceSession(subfmodel.serialized).run(None, ffeed)[0]
        qoutputs = onnxruntime.InferenceSession(subqmodel.serialized).run(None, qfeed)[0]
        scale, _ = _get_scale_zp(qnode, qmodel)

        # Compare results
        qerror = eval_metrics(compare_outputs(foutputs, qoutputs, scale, per_channel=per_channel))
        summary[f"{qnode.name} ({qnode.op_type})"] = qerror
    return summary


def cumulative_quantization_error(fmodel, qmodel, target_node=None, batch_size=16, seed=None):
    """Measures the cumulative quantization error in a set of ONNX models

    Args:
        fmodel (onnx.ModelProto): the float model.
        qmodel (onnx.ModelProto): the quantized version of `fmodel`.
        target_node (str, optional): error computation is performed only in the target node,
            expanding the analysis to each output channel. Defaults to None.
        batch_size (int, optional): the batch size. Defaults to 16.
        seed (int, optional): a random seed. Defaults to None.

    Returns:
        dict: the quantization error by each layer
    """
    def _copy_vi_and_replace_type(x, new_type):
        new_vi = x.__deepcopy__()
        new_vi.type.tensor_type.elem_type = new_type
        return new_vi

    # We clone the models to avoid modifying the original ones
    qmodel = ONNXModel(qmodel).clone()
    fmodel = ONNXModel(fmodel).clone()

    # Input name between both models should be the same
    in_qnames = [vi.name for vi in qmodel.input]
    if [vi.name for vi in fmodel.input] != in_qnames:
        raise RuntimeError(f"{fmodel.name} does not contain tensor with names {in_qnames}")

    # Search target nodes and update graph outputs
    per_channel = target_node is not None
    qtarget_nodes = _search_quantized_target_nodes(qmodel, target_node_name=target_node)

    # Create the intermediate models with multiple outputs
    fout_type = fmodel.output[0].type.tensor_type.elem_type
    fmodel.graph().ClearField("output")
    qmodel.graph().ClearField("output")
    foutput_names = [node.output[0] for node in fmodel.nodes()]
    for qnode in qtarget_nodes:
        # Search qnode.output in ValueInfoProto.
        # It will be a new output of the quantized model.
        qvi = qmodel.find_value_info_by_name(qnode.output[0])
        qmodel.output.append(qvi)

        # Intermediate quantized outputs must be contained in fmodel
        if qvi.name not in foutput_names:
            raise RuntimeError(f"{fmodel.name} does not contain '{qvi.name}'")

        # Since the name of the intermediate outputs between fmodel and qmodel must be the same,
        # we can include qvi as an output of fmodel. Note that the type must be updated
        vi = _copy_vi_and_replace_type(qvi, new_type=fout_type)
        fmodel.output.append(vi)

    # Generate a random set of samples and split them in batches of 1
    samples = generate_onnx_random_samples(fmodel, batch_size=batch_size, seed=seed)

    # Create sessions
    fsess = onnxruntime.InferenceSession(fmodel.serialized)
    qsess = onnxruntime.InferenceSession(qmodel.serialized)

    # Compute cumulative quantization error
    summary = {}
    outputs = fsess.run(None, samples), qsess.run(None, samples)
    for qnode, foutputs, qoutputs in zip(qtarget_nodes, *outputs):
        key = f"{qnode.name} ({qnode.op_type})"
        # Retrieve scale to dequantize qoutputs
        qscales, _ = _get_scale_zp(qnode, qmodel)

        # Compute cumulative quantization error
        qerror = eval_metrics(compare_outputs(foutputs, qoutputs, qscales, per_channel=per_channel))
        summary[key] = qerror
    return summary
